import unittest
from typing import Optional
from unittest.mock import patch

from parameterized import parameterized  # type: ignore

from vmngclient.session import vManageSession


class TestSession(unittest.TestCase):
    def setUp(self):
        self.url = "example.com"
        self.username = "admin"
        self.password = "admin_password"

    def test_session_str(self):
        # Arrange, Act
        session = vManageSession(self.url, self.username, self.password, port=111)

        # Assert
        self.assertEqual(str(session), "admin@https://example.com:111")

    @parameterized.expand(
        [
            (123, "https://example.com:123"),
            (None, "https://example.com"),
        ]
    )
    def test_base_url(self, port: Optional[int], base_url: str):
        # Arrange, Act
        session = vManageSession(self.url, self.username, self.password, port=port)

        # Assert
        self.assertEqual(session.base_url, base_url)

    def test_session_repr(self):
        # Arrange, Act
        session = vManageSession("domain.com", "user1", "$password", port=111)
        session_str = "vManageSession('domain.com', 'user1', '$password', port=111, subdomain='None')"
        # Assert
        self.assertEqual(repr(session), session_str)

    def test_session_repr_different_sessions(self):
        # Arrange, Act
        session = vManageSession("domain.com", "user1", "$password", port=111)
        session_str = "vManageSession('not.domain.com', 'different_user', '$password', port=111, subdomain='None')"
        # Assert
        self.assertNotEqual(repr(session), session_str)

    @patch("vmngclient.session.Session.__repr__")
    def test_session_eval_repr(self, mock_repr):
        # Arrange, Act
        mock_repr.return_value = "vManageSession('domain.com', 'user1', '$password', port=111, subdomain='None')"
        session = vManageSession("domain.com", "user1", "$password", port=111)
        # Assert
        self.assertEqual(eval(mock_repr()), session)

    @patch("vmngclient.session.Session.__repr__")
    def test_session_eval_repr_different_sessions(self, mock_repr):
        # Arrange, Act
        mock_repr.return_value = "vManageSession('domain.com', 'user1', '$password', port=111, subdomain='None')"
        session = vManageSession("not.domain.com", "different_user", "$password", port=111)
        # Assert
        self.assertNotEqual(eval(mock_repr()), session)

    def test_session_eq(self):
        # Arrange, Act
        session_1 = vManageSession("domain.com", "user1", "$password", port=111)
        session_2 = vManageSession("domain.com", "user1", "$password", port=111)
        # Assert
        self.assertEqual(session_1, session_2)

    def test_session_eq_different_sessions(self):
        # Arrange, Act
        session_1 = vManageSession("domain.com", "user1", "$password", port=111)
        session_2 = vManageSession("not.domain.com", "different_user", "$password", port=111)
        # Assert
        self.assertNotEqual(session_1, session_2)

    @parameterized.expand(
        [
            (123, "/devices", "https://example.com:123/devices"),
            (123, "devices", "https://example.com:123/devices"),
            (None, "/devices", "https://example.com/devices"),
            (None, "devices", "https://example.com/devices"),
        ]
    )
    def test_get_full_url(self, port: Optional[int], url: str, full_url: str):
        # Arrange, Act
        session = vManageSession(self.url, self.username, self.password, port=port)

        # Assert
        self.assertEqual(session.get_full_url(url), full_url)


if __name__ == '__main__':
    unittest.main()



import io
import tempfile
import unittest
from pathlib import Path
from unittest.mock import ANY, patch

from vmngclient.api.admin_tech_api import (
    AdminTechAPI,
    DownloadAdminTechLogError,
    GenerateAdminTechLogError,
    RequestTokenIdNotFound,
)
from vmngclient.dataclasses import DeviceAdminTech


class TestAdminTechAPI(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.device_ip = "169.254.10.4"
        self.admin_tech_generate_response = {
            "size": 1479740,
            "fileName": "172.16.253.129-vm129-20221116-065219-admin-tech.tar.gz",
        }
        self.device_admin_tech_infos = {
            "data": [
                {
                    "fileName": "172.16.255.200-vm200-20221116-060922-admin-tech.tar.gz",
                    "creationTime": 1668578961380,
                    "size": 31934859,
                    "state": "done",
                    "requestTokenId": "null",
                }
            ]
        }
        self.admin_tech_infos = {
            "data": [
                {
                    "creationTime": 1668581537503,
                    "size": self.admin_tech_generate_response["size"],
                    "fileName": self.admin_tech_generate_response["fileName"],
                    "state": "done",
                    "tac_state": "notStarted",
                    "deviceIP": self.device_ip,
                    "local-system-ip": "172.16.253.129",
                    "requestTokenId": "aace1605-ba9a-40fa-990b-c694a011a866",
                },
                {
                    "creationTime": 1668578961380,
                    "size": 31934859,
                    "fileName": "172.16.255.200-vm200-20221116-060922-admin-tech.tar.gz",
                    "state": "done",
                    "tac_state": "notStarted",
                    "deviceIP": "169.254.10.1",
                    "local-system-ip": "172.16.255.200",
                    "requestTokenId": "8b7f9a3a-e137-4af4-b2f9-b04808e1e2eb",
                },
            ]
        }
        self.admin_tech_info = self.admin_tech_infos["data"][0]
        self.download_file_content = "Downloaded file content"
        self.download_file = io.BytesIO(self.download_file_content.encode())

    @patch("vmngclient.session.vManageSession")
    @patch("requests.Response")
    def test_get(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.json.return_value = self.device_admin_tech_infos
        # Act
        admintechs = AdminTechAPI(mock_session).get(self.device_ip)
        # Assert
        mock_session.post.assert_called_once_with(
            url="/dataservice/device/tools/admintechlist", json={"deviceIP": self.device_ip}
        )
        self.assertIsInstance(admintechs[0], DeviceAdminTech)

    @patch("vmngclient.session.vManageSession")
    @patch("requests.Response")
    def test_get_all(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_response.json.return_value = self.admin_tech_infos
        # Act
        admintechs = AdminTechAPI(mock_session).get_all()
        # Assert
        mock_session.get.assert_called_once_with("/dataservice/device/tools/admintechs")
        self.assertEqual(len(admintechs), len(self.admin_tech_infos["data"]))

    @patch("vmngclient.session.vManageSession")
    @patch("requests.Response")
    def test_generate(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = 200
        mock_response.json.return_value = self.admin_tech_generate_response
        # Act
        filename = AdminTechAPI(mock_session).generate(
            device_id=self.device_ip, polling_timeout=0.01, polling_interval=0.01
        )
        print(filename)
        # Assert
        mock_session.post.assert_called_once_with(url="/dataservice/device/tools/admintech", json=ANY, timeout=ANY)
        self.assertEqual(filename, self.admin_tech_generate_response["fileName"])

    @patch("vmngclient.session.vManageSession")
    @patch("requests.Response")
    def test_generate_in_progress_error_retry(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = 400
        mock_response.json.return_value = {"error": {"details": "Admin tech creation already in progress"}}
        interval = 0.01
        count = 2
        # Act/Assert
        with self.assertRaises(GenerateAdminTechLogError):
            AdminTechAPI(mock_session).generate(
                device_id=self.device_ip, polling_timeout=interval * count, polling_interval=interval
            )
        self.assertEqual(mock_session.post.call_count, count)

    @patch("vmngclient.session.vManageSession")
    @patch("requests.Response")
    def test_generate_error(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = 500
        mock_response.json.return_value = {"error": {"details": "Server Error"}}
        interval = 0.01
        count = 3
        # Act/Assert
        with self.assertRaises(GenerateAdminTechLogError):
            AdminTechAPI(mock_session).generate(
                device_id=self.device_ip, polling_timeout=interval * count, polling_interval=interval
            )
        mock_session.post.assert_called_once()

    @patch("vmngclient.session.vManageSession")
    @patch("requests.Response")
    def test_delete(self, mock_session, mock_response):
        # Arrange
        filename = self.admin_tech_generate_response["fileName"]
        token_id = self.admin_tech_info["requestTokenId"]
        mock_session.get.return_value = mock_response
        mock_response.json.return_value = self.admin_tech_infos
        # Act
        AdminTechAPI(mock_session).delete(filename)
        # Assert
        mock_session.delete.assert_called_once_with(f"/dataservice/device/tools/admintech/{token_id}")

    @patch("vmngclient.session.vManageSession")
    @patch("requests.Response")
    def test_delete_token_not_found(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_response.json.return_value = self.admin_tech_infos
        # Act/Assert
        with self.assertRaises(RequestTokenIdNotFound):
            AdminTechAPI(mock_session).delete("fake-filename.tar.gz")

    @patch("vmngclient.session.vManageSession")
    @patch("requests.Response")
    def test_download(self, mock_session, mock_response):
        # Arrange
        filename = self.admin_tech_generate_response["fileName"]
        mock_session.get_file.return_value = mock_response
        mock_session.get.return_value = mock_response
        mock_response.status_code = 200
        mock_response.content = self.download_file_content
        with tempfile.TemporaryDirectory() as tmpdir:
            # Act
            download_path = AdminTechAPI(mock_session).download(filename, Path(tmpdir))
            # Assert
            self.assertEqual(download_path, Path(tmpdir) / filename)

    @patch("vmngclient.session.vManageSession")
    @patch("requests.Response")
    def test_download_error(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_response.status_code = 500
        mock_response.json.return_value = {"error": {"details": "Server Error"}}
        with tempfile.TemporaryDirectory() as tmpdir:
            # Act/Assert
            with self.assertRaises(DownloadAdminTechLogError):
                AdminTechAPI(mock_session).download("fake-filename.tar.gz", Path(tmpdir))



import json
import logging
from difflib import Differ
from enum import Enum
from typing import List, cast

from ciscoconfparse import CiscoConfParse  # type: ignore
from requests.exceptions import HTTPError
from tenacity import retry, retry_if_result, stop_after_attempt, wait_fixed  # type: ignore

from vmngclient.dataclasses import Device, Template
from vmngclient.session import vManageSession
from vmngclient.utils.creation_tools import create_dataclass
from vmngclient.utils.device_model import DeviceModel
from vmngclient.utils.operation_status import OperationStatus

logger = logging.getLogger(__name__)


class TemplateType(Enum):
    CLI = 'file'
    FEATURE = 'template'


class NotFoundError(Exception):
    """Used when a template item is not found."""

    def __init__(self, template):
        self.message = f"No such template: '{template}'"


class NameAlreadyExistError(Exception):
    """Used when a template item exists."""

    def __init__(self, name):
        self.message = f"Template with that name '{name}' exists."


class AttachedError(Exception):
    """Used when delete attached template."""

    def __init__(self, template):
        self.message = f"Template: {template} is attached to device."


class TemplateTypeError(Exception):
    """Used when wrong type template."""

    def __init__(self, name):
        self.message = f"Template: {name} - wrong template type."


class TemplateAPI:
    def __init__(self, session: vManageSession) -> None:
        self.session = session

    @property
    def templates(self) -> List[Template]:
        """

        Returns:
            List[Template]: List of existing templates.
        """
        endpoint = "/dataservice/template/device"
        data = self.session.get_data(endpoint)
        return [create_dataclass(Template, template) for template in data]

    def get(self, name: str) -> Template:
        """

        Args:
            name (str): Name of template.

        Raises:
            NotFoundError: If template does not exist.

        Returns:
            Template: Selected template.
        """
        for template in self.templates:
            if name == template.name:
                return template
        raise NotFoundError(name)

    def get_id(self, name: str) -> str:
        """

        Args:
            name (str): Name of template to get id.

        Returns:
            str: Template id.
        """
        return self.get(name).id

    def wait_for_complete(self, operation_id: str, timeout_seconds: int = 300, sleep_seconds: int = 5) -> bool:
        """

        Args:
            operation_id (str): Id of the action waiting for completion.
            timeout_seconds (int, optional): Time for completion. Defaults to 300.
            sleep_seconds (int, optional): Sleep time between repetitions. Defaults to 5.

        Returns:
            bool: True if acction status is successful, otherwise - False.
        """

        def _log_exception(retry_state):
            logger.error(
                f"Operation status not achieved in the given time, exception: {retry_state.outcome.exception()}."
            )
            return False

        def check_status(action_data):
            if action_data:
                list_action = [action == OperationStatus.SUCCESS for action in action_data]
                return not all(list_action)
            else:
                return True

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_status),
            retry_error_callback=_log_exception,
        )
        def wait_for_status():
            return self.get_operation_status(operation_id)

        if wait_for_status():
            logger.info(f"The action: {operation_id} - successful")
            return True
        else:
            logger.info(f"The action: {operation_id} - failed")
            return False

    def attach(self, name: str, device: Device) -> bool:
        """

        Args:
            name (str): Template name to attached.
            device (Device): Device to attach template.

        Returns:
            bool: True if attaching template is successful, otherwise - False.
        """
        try:
            template_id = self.get_id(name)
            self.template_validation(template_id, device=device)
        except NotFoundError:
            logger.error(f"Error, Template with name {name} not found on {device}.")
            return False
        except HTTPError as error:
            error_details = json.loads(error.response.text)
            logger.error(f"Error in config: {error_details['error']['details']}.")
            return False
        payload = {
            "deviceTemplateList": [
                {
                    "templateId": template_id,
                    "device": [
                        {
                            "csv-status": "complete",
                            "csv-deviceId": device.uuid,
                            "csv-deviceIP": device.id,
                            "csv-host-name": device.hostname,
                            "csv-templateId": template_id,
                        }
                    ],
                }
            ]
        }
        endpoint = "/dataservice/template/device/config/attachcli"
        response = self.session.post(url=endpoint, json=payload).json()
        logger.info(f"Attaching a template: {name} to the device: {device.hostname}.")
        return self.wait_for_complete(response['id'])

    def device_to_cli(self, device: Device) -> bool:
        """

        Args:
            device (Device): Device to chcange mode.

        Returns:
            bool: True if change mode to cli is successful, otherwise - False.
        """
        payload = {
            "deviceType": device.personality.value,
            "devices": [{"deviceId": device.uuid, "deviceIP": device.id}],
        }
        endpoint = "/dataservice/template/config/device/mode/cli"
        response = self.session.post(url=endpoint, json=payload).json()
        logger.info(f"Changing mode to cli mode for {device.hostname}.")
        return self.wait_for_complete(response['id'])

    def get_operation_status(self, operation_id: str) -> List[OperationStatus]:
        """

        Args:
            operation_id (str): Operation id.

        Returns:
            List[OperationStatus]: List of operations.
        """
        endpoint = f"/dataservice/device/action/status/{operation_id}"
        response = cast(list, self.session.get_data(endpoint))
        return [OperationStatus(status['status']) for status in response]

    def delete(self, name: str) -> bool:
        """

        Args:
            name (str): Name template to delete.

        Raises:
            AttachedError: If template is attached to device.

        Returns:
            bool: True if deletion is successful, otherwise - False.
        """
        template = self.get(name)
        endpoint = f"/dataservice/template/device/{template.id}"
        if template.devices_attached == 0:
            response = self.session.delete(url=endpoint)
            logger.info(f"Template with name: {name} - deleted.")
            return response.ok
        logger.info(f"Template: {template} is attached to device - cannot be deleted.")
        raise AttachedError(template.name)

    def create(self, device_model: DeviceModel, name: str, description: str, config: CiscoConfParse) -> bool:
        """

        Args:
            device_model (DeviceModel): Device model to create template.
            name (str): Name template to create.
            description (str): Description template to create.
            config (CiscoConfParse): The config to device.

        Raises:
            NameAlreadyExistError: If such template name already exists.

        Returns:
            bool: True if create template is successful, otherwise - False.
        """
        try:
            self.get(name)
            logger.error(f"Error, Template with name: {name} exists.")
            raise NameAlreadyExistError(name)
        except NotFoundError:
            cli_template = CLITemplate(self.session, device_model, name, description)
            cli_template.config = config
            logger.info(f"Template with name: {name} - created.")
            return cli_template.send_to_device()

    def template_validation(self, id: str, device: Device) -> str:
        """Checking the template of the configuration on the machine.

        Args:
            id (str): template id to check.
            device (Device): The device on which the configuration is to be validate.

        Returns:
            str: Validated config.
        """
        payload = {
            "templateId": id,
            "device": {
                "csv-status": "complete",
                "csv-deviceId": device.uuid,
                "csv-deviceIP": device.id,
                "csv-host-name": device.hostname,
                "csv-templateId": id,
            },
            "isEdited": False,
            "isMasterEdited": False,
            "isRFSRequired": True,
        }
        endpoint = "/dataservice/template/device/config/config/"
        response = self.session.post(url=endpoint, json=payload)
        return response.text

    @staticmethod
    def compare_template(first: CiscoConfParse, second: CiscoConfParse, full: bool = False, debug: bool = False) -> str:
        """

        Args:
            first: First template for comparison.
            second: Second template for comparison.
            full: Return a full comparison if True, otherwise only the lines that differ.
            debug: Adding debug to the logger. Defaults to False.

        Returns:
            str: The compared templates.

        Code    Meaning
        '- '    line unique to sequence 1
        '+ '    line unique to sequence 2
        '  '    line common to both sequences
        '? '    line not present in either input sequence

        Example:
        >>> a = "!\n  tacacs\n  server 192.168.1.1\n   vpn 2\n   secret-key a\n   auth-port 151\n exit".splitlines()
        >>> b = "!\n  tacacs\n  server 192.168.1.1\n   vpn 3\n   secret-key a\n   auth-port 151\n exit".splitlines()
        >>> a_conf = CiscoConfParse(a)
        >>> b_conf = CiscoConfParse(b)
        >>> compare = TemplateAPI.compare_template(a_conf, b_conf full=True)
        >>> print(compare)
          !
            tacacs
            server 192.168.1.1
        -    vpn 2
        ?        ^
        +    vpn 3
        ?        ^
            secret-key a
            auth-port 151
        exit
        """
        first_n = list(map(lambda x: x + "\n", first.ioscfg))
        second_n = list(map(lambda x: x + "\n", second.ioscfg))
        compare = list(Differ().compare(first_n, second_n))
        if not full:
            compare = [x for x in compare if x[0] in ["?", "-", "+"]]
        if debug:
            logger.debug("".join(compare))
        return "".join(compare)

    def compare_with_running(
        self, template: CiscoConfParse, device: Device, full: bool = False, debug: bool = False
    ) -> str:
        """The comparison of the config with the one running on the machine.

        Args:
            template: The template to compare.
            device: The device on which to compare config.
            full: Return a full comparison if True, otherwise only the lines that differ.
            debug: Adding debug to the logger. Defaults to False.

        Returns:
            str: The compared templates.

        Example:
        >>> a = "!\n  tacacs\n  server 192.168.1.1\n   vpn 512\n   secret-key a\n   auth-port 151\n exit".splitlines()
        >>> a_conf = CiscoConfParse(a)
        >>> device = DevicesAPI(API_SESSION).get(DeviceField.HOSTNAME, device_name)
        >>> compare = TemplateAPI.compare_template(a_conf, device, full=True)
        >>> print(compare)
        .
        .
        .
            zbfw-udp-idle-time    30
           !
          !
        + !
        +   tacacs
        +   server 192.168.1.1
        +    vpn vpn 512
        +    secret-key a
        +    auth-port 151
        +  exit
          omp
           no shutdown
           ecmp-limit       6
        .
        .
        .
        """
        running_config = CLITemplate(
            self.session, DeviceModel(device.model), 'running_conf', 'running_conf'
        ).load_running(device)
        return self.compare_template(running_config, template, debug)


class CLITemplate:
    def __init__(self, session: vManageSession, device_model: DeviceModel, name: str, description: str) -> None:
        self.session = session
        self.device_model = device_model
        self.name = name
        self.description = description
        self.config: CiscoConfParse = CiscoConfParse([])

    def load(self, id: str) -> CiscoConfParse:
        """Load CLI config from template.

        Args:
            id (str): The template id from which load config.

        Raises:
            TemplateTypeError: wrong template type - CLI required.

        Returns:
            CiscoConfParse: Loaded template.
        """
        endpoint = f"/dataservice/template/device/object/{id}"
        config = self.session.get_json(endpoint)
        if TemplateType(config['configType']) == TemplateType.FEATURE:
            raise TemplateTypeError(config['templateName'])
        self.config = CiscoConfParse(config['templateConfiguration'].splitlines())
        return self.config

    def load_running(self, device: Device) -> CiscoConfParse:
        """Load running config from device.

        Args:
            device (Device): The device from which load config.

        Returns:
            CiscoConfParse: A working configuration on the machine.
        """
        endpoint = f"/dataservice/template/config/running/{device.uuid}"
        config = self.session.get_json(endpoint)
        self.config = CiscoConfParse(config['config'].splitlines())
        logger.debug(f"Template loaded from {device.hostname}.")
        return self.config

    def send_to_device(self) -> bool:
        """

        Returns:
            bool: True if send template to device is successful, otherwise - False.

        The payload differs depending on the type of machine - for physical machines it has two more attributes.
        """
        config_str = "\n".join(self.config.ioscfg)
        payload = {
            "templateName": self.name,
            "templateDescription": self.description,
            "deviceType": self.device_model.value,
            "templateConfiguration": config_str,
            "factoryDefault": False,
            "configType": "file",
        }
        if self.device_model not in [DeviceModel.VEDGE, DeviceModel.VSMART, DeviceModel.VMANAGE, DeviceModel.VBOND]:
            payload["cliType"] = "device"
            payload["draftMode"] = False

        endpoint = "/dataservice/template/device/cli/"
        try:
            self.session.post(url=endpoint, json=payload).json()
        except HTTPError as error:
            response = json.loads(error.response.text)['error']
            logger.error(response['message'])
            logger.error(response['details'])
            return False
        logger.info(f"Template with name: {self.name} - sent to the device.")
        return True

    def update(self, id: str) -> bool:
        """

        Args:
            id (str): Template id to update.

        Returns:
            bool: True if update template is successful, otherwise - False.
        """
        config_str = "\n".join(self.config.ioscfg)
        payload = {
            "templateId": id,
            "templateName": self.name,
            "templateDescription": self.description,
            "deviceType": self.device_model.value,
            "templateConfiguration": config_str,
            "factoryDefault": False,
            "configType": "file",
            "draftMode": False,
        }
        endpoint = f"/dataservice/template/device/{id}"
        try:
            self.session.put(url=endpoint, json=payload)
        except HTTPError as error:
            response = json.loads(error.response.text)['error']
            logger.error(response['message'])
            logger.error(response['details'])
            return False
        logger.info(f"Template with name: {self.name} - updated.")
        return True

    def add_to_config(self, add_config: CiscoConfParse, add_before: str) -> None:
        """Add config to existing config before provided value.

        Args:
            add_config (CiscoConfParse): Config to added.
            add_before (str): The value before which to add config.
        """
        for comand in add_config.ioscfg:
            self.config.ConfigObjs.insert_before(add_before, comand, atomic=True)



from __future__ import annotations

import logging
import time
from enum import Enum, auto
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urljoin

from requests import Response, Session
from requests.auth import AuthBase
from requests.exceptions import HTTPError
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed  # type: ignore

from vmngclient.utils.response import response_debug
from vmngclient.vmanage_auth import vManageAuth

logger = logging.getLogger(__name__)

JSON = Union[Dict[str, "JSON"], List["JSON"], str, int, float, bool, None]


class SessionType(Enum):
    PROVIDER = auto()
    TENANT = auto()
    PROVIDER_AS_TENANT = auto()
    NOT_DEFINED = auto()


class UserMode(Enum):
    PROVIDER = "provider"
    TENANT = "tenant"
    NOT_RECOGNIZED = "not recognized"
    NOT_FOUND = "not found"


class ViewMode(Enum):
    PROVIDER = "provider"
    TENANT = "tenant"
    NOT_RECOGNIZED = "not recognized"
    NOT_FOUND = "not found"


class SessionNotCreatedError(Exception):
    pass


def create_vManageSession(
    url: str,
    username: str,
    password: str,
    port: Optional[int] = None,
    subdomain: Optional[str] = None,
) -> vManageSession:
    """Factory function that creates session object based on provided arguments.

    Args:
        url (str): IP address or domain name
        port (int): port
        username (str): username
        password (str): password
        subdomain: subdomain specifying to which view switch when creating provider as a tenant session,
            works only on provider user mode

    Returns:
        Session object

    """
    session = vManageSession(url=url, username=username, password=password, port=port, subdomain=subdomain)
    session.auth = vManageAuth(session.base_url, username, password, verify=False)
    if subdomain:
        tenant_id = session.get_tenant_id()
        vsession_id = session.get_virtual_session_id(tenant_id)
        session.headers.update({'VSessionId': vsession_id})
    server_info = session.server()

    try:
        user_mode = UserMode(server_info.get("userMode", "not found"))
    except ValueError:
        user_mode = UserMode.NOT_RECOGNIZED
        logger.warning(f"Unrecognized user mode is: '{server_info.get('userMode')}'")

    try:
        view_mode = ViewMode(server_info.get("viewMode", "not found"))
    except ValueError:
        view_mode = ViewMode.NOT_RECOGNIZED
        logger.warning(f"Unrecognized user mode is: '{server_info.get('viewMode')}'")
    if user_mode is UserMode.TENANT and not subdomain and view_mode is ViewMode.TENANT:
        session.session_type = SessionType.TENANT
    elif user_mode is UserMode.PROVIDER and not subdomain and view_mode is ViewMode.PROVIDER:
        session.session_type = SessionType.PROVIDER
    elif user_mode is UserMode.PROVIDER and view_mode is ViewMode.TENANT:

        session.session_type = SessionType.PROVIDER_AS_TENANT
    elif user_mode is UserMode.TENANT and subdomain:
        raise SessionNotCreatedError(
            f"Session not created. Subdomain {subdomain} passed to tenant session, "
            "cannot switch to tenant from tenant user mode."
        )
    else:
        session.session_type = SessionType.NOT_DEFINED
        logger.warning(
            f"Session created with {user_mode.value} user mode and {view_mode.value} view mode.\n"
            f"Session type set to not defined"
        )
    logger.info(f"Logged as {username}. The session type is {session.session_type}")
    return session


class vManageSession(Session):
    """Base class for API sessions for vManage client.

    Defines methods and handles session connectivity available for provider, provider as tenant, and tenant.

    Args:
        url: IP address or domain name, i.e. '10.0.1.200' or 'example.com'
        port: port
        username: username
        password: password
    """

    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        verify: bool = False,
        port: Optional[int] = None,
        subdomain: Optional[str] = None,
        auth: Optional[AuthBase] = None,
    ):
        self.url = url
        self.port = port
        self.base_url = self.__create_base_url()
        self.username = username
        self.password = password
        self.port = port
        self.subdomain = subdomain

        self.session_type = SessionType.NOT_DEFINED

        super(vManageSession, self).__init__()
        self.__prepare_session(verify, auth)

    def request(self, method, url, *args, **kwargs) -> Any:
        full_url = self.get_full_url(url)
        response = super(vManageSession, self).request(method, full_url, *args, **kwargs)
        logger.debug(response_debug(response))
        try:
            response.raise_for_status()
        except HTTPError as error:
            logger.debug(error)
            if response.status_code == 403:
                logger.info(f"User {self.username} is unauthorized for method {method} {full_url}")
            else:
                raise error
        return response

    def get_full_url(self, url_path: str) -> str:
        """Returns base API url plus given url path."""
        return urljoin(self.base_url, url_path)

    def __create_base_url(self) -> str:
        """Creates base url based on ip address or domain and port if provided.

        Returns:
            str: Base url shared for every request.
        """
        if self.port:
            return f"https://{self.url}:{self.port}"
        return f"https://{self.url}"

    def about(self) -> Dict:
        return self.get_data(url="/dataservice/client/about")

    def server(self) -> Dict:
        return self.get_data(url="/dataservice/client/server")

    def get_data(self, url: str) -> Any:
        return self.get_json(url)['data']

    def get_json(self, url: str) -> Any:
        response = self.get(url)
        return response.json()

    def get_file(self, url: str, filename: Path) -> Response:
        """Get a file using session get.

        Args:
            url: dataservice api.
            filename: Filename to write download file to.

        Returns:
            http response.

        Example usage:
            response = self.session.get_file(url, filename)

        """
        with self.get(url) as response:
            with open(filename, "wb") as file:
                file.write(response.content)
        return response

    def wait_for_server_reachability(self, retries: int, delay: int, initial_delay: int = 0) -> bool:
        """Checks if vManage API is reachable by sending server request.

        Retries on HTTPError for specified number of times.
        Delays between each request are configurable,
        It is intended to be used as a probe, so it doesn't raise original error from exception.

        Args:
            retries: total number of retires
            delay: time to wait between each retry
            initial_delay: time before sending first request

        Returns:
            Bool: True if device is reachable, False if not
        """

        def _log_exception(retry_state):
            logger.error(f"Cannot reach server, original exception: {retry_state.outcome.exception()}")
            return False

        if initial_delay:
            time.sleep(initial_delay)

        @retry(
            wait=wait_fixed(delay),
            retry=retry_if_exception_type(HTTPError),
            stop=stop_after_attempt(retries),
            retry_error_callback=_log_exception,
        )
        def _send_server_request():
            return self.server()

        return True if _send_server_request() else False

    def get_tenant_id(self) -> str:
        """Gets tenant UUID for its subdomain.

        Returns:
            Tenant UUID.
        """
        tenants = self.get_data(url='/dataservice/tenant')
        tenant_id = [tenant.get('tenantId', None) for tenant in tenants if tenant['subDomain'] == self.subdomain][0]
        return tenant_id

    def get_virtual_session_id(self, tenant_id: str) -> str:
        """Get VSessionId for a specific tenant

        Note: In a multitenant vManage system, this API is only available in the Provider view.

        Args:
            tenant_id: provider or tenant UUID
        Returns:
            Virtual session token
        """
        url_path = f"/dataservice/tenant/{tenant_id}/vsessionid"
        response = self.post(url_path)
        return response.json()['VSessionId']

    def __prepare_session(self, verify: bool, auth: Optional[AuthBase]) -> None:
        self.auth = auth
        self.verify = verify

    def __str__(self) -> str:
        return f"{self.username}@{self.base_url}"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}('{self.url}', '{self.username}', '{self.password}', port={self.port}, "
            f"subdomain='{self.subdomain}')"
        )

    def __eq__(self, other):
        if isinstance(other, vManageSession):
            comparison_list = [
                self.url == other.url,
                self.username == other.username,
                self.password == other.password,
                self.port == other.port,
                str(self.subdomain) == str(other.subdomain),
            ]
            return True if all(comparison_list) else False
        return False



from dataclasses import dataclass
from pprint import pformat
from typing import Any, Dict, List, Type, TypeVar, cast

from requests import Response
from requests.exceptions import JSONDecodeError

from vmngclient.utils.creation_tools import create_dataclass

T = TypeVar("T")


class VManageResponseException(Exception):
    def __init__(self, message: str, response: Response):
        self.message = message
        self.response = response

    def __str__(self):
        return f"{self.message}\n{response_debug(self.response)}"


@dataclass
class VManageResponseDebugInfo:
    request: Dict[str, Any]
    response: Dict[str, Any]


def response_debug(response: Response, headers: bool = False) -> str:
    request_body = response.request.body
    if isinstance(request_body, bytes):
        request_body = str(request_body, encoding="utf-8")
    info = VManageResponseDebugInfo(
        request={"method": response.request.method, "url": response.request.url, "body": request_body},
        response={"status": response.status_code, "reason": response.reason},
    )
    if len(response.text) <= 1024:
        info.response.update({"text": response.text})
    else:
        info.response.update({"text(trimmed)": response.text[:128]})
    if headers:
        info.request.update({"headers": dict(response.request.headers.items())})
        info.response.update({"headers": dict(response.headers.items())})
    return pformat(dict(info.__dict__.items()), width=80, indent=0)


def get_json_data(response: Response) -> Any:
    try:
        response_json = response.json()
    except JSONDecodeError:
        raise VManageResponseException("Response does not contain a valid json", response)
    if isinstance(response_json, dict) and "data" in response_json.keys():
        return response_json.get("data")
    else:
        raise VManageResponseException("{\"data\": ... } field not found in Response", response)


def get_json_data_as_list(response: Response) -> List:
    data = get_json_data(response)
    if not isinstance(data, list):
        raise VManageResponseException(f"{{\"data\": ...}} contains {type(data)}, expected list", response)
    return cast(list, data)


def get_json_data_as_dict(response: Response) -> Dict[str, Any]:
    data = get_json_data(response)
    if not isinstance(data, Dict):
        raise VManageResponseException(f"{{\"data\": ...}} contains {type(data)}, expected dict", response)
    return cast(dict, data)


def parse_as(response: Response, cls: Type[T]) -> T:
    try:
        return create_dataclass(cls, get_json_data_as_dict(response))
    except TypeError as e:
        raise VManageResponseException(f"Cannot create {cls}: {str(e)}", response)


def parse_as_list(response: Response, cls: Type[T]) -> List[T]:
    try:
        items = get_json_data_as_list(response)
        return [create_dataclass(cls, item) for item in items]
    except TypeError as e:
        raise VManageResponseException(f"Cannot create {cls}: {str(e)}", response)



"""
Module for handling admintech logs for a device
"""
import logging
import time
from pathlib import Path
from typing import List, Optional

from requests import Response
from requests.exceptions import HTTPError

from vmngclient.dataclasses import AdminTech, DeviceAdminTech
from vmngclient.session import vManageSession
from vmngclient.utils.creation_tools import create_dataclass

logger = logging.getLogger(__name__)


class GenerateAdminTechLogError(Exception):
    pass


class DownloadAdminTechLogError(Exception):
    pass


class RequestTokenIdNotFound(Exception):
    pass


class AdminTechAPI:
    """Class for handling admintech logs for a device.

    Attributes:
        session: logged in API client session
    """

    def __init__(self, session: vManageSession) -> None:
        self.session = session

    def __str__(self) -> str:
        return str(self.session)

    def get(self, device_id: str) -> List[DeviceAdminTech]:
        """Gets admintech log information for a device.

        Args:
            device_id: device ID (usually system-ip)
        Returns:
            AdminTech object list for given device
        """
        body = {'deviceIP': device_id}
        response = self.session.post(url='/dataservice/device/tools/admintechlist', json=body)
        items = response.json()["data"]
        return [create_dataclass(DeviceAdminTech, item) for item in items]

    def get_all(self) -> List[AdminTech]:
        """Gets admintech log information for all devices.

        Returns:
            AdminTech objects list for all devices
        """
        response = self.session.get('/dataservice/device/tools/admintechs')
        items = response.json()["data"]
        return [create_dataclass(AdminTech, item) for item in items]

    def generate(
        self,
        device_id: str,
        exclude_cores: bool = True,
        exclude_tech: bool = False,
        exclude_logs: bool = True,
        request_timeout: int = 3600,
        polling_timeout: int = 1200,
        polling_interval: int = 30,
    ) -> str:
        """Generates admintech log for a device.
        Args:
            device_id: device ID (usually system-ip)
            exclude_cores: exclude core in generated admintech log file
            exclude_tech: exclude tech in generated admintech log file
            exclude_logs: exclude logs in generated admintech log file
            request_timeout: wait time in seconds to generate admintech after request
            polling_timeout: retry period in seconds for successfull request
            polling_interval: polling interval in seconds between request attempts
        Returns:
            filename of generated admintech log
        """
        create_admin_tech_error_msgs = "Admin tech creation already in progress"
        body = {
            "deviceIP": device_id,
            "exclude-cores": exclude_cores,
            "exclude-tech": exclude_tech,
            "exclude-logs": exclude_logs,
        }
        polling_timer = polling_timeout
        while polling_timer > 0:
            logger.info(
                f"Starting AdminTech log creation for {device_id}, waiting up to {request_timeout} seconds to complete"
            )
            try:
                response = self.session.post(
                    url="/dataservice/device/tools/admintech", json=body, timeout=request_timeout
                )
            except HTTPError as http_error:
                response = http_error.response
            if response.status_code == 200:
                return response.json()["fileName"]
            if response.status_code == 400 and create_admin_tech_error_msgs in response.json().get("error", {}).get(
                "details", ""
            ):
                logger.warning(f"Admin tech creation already in progress, retrying in {polling_interval} seconds")
            else:
                raise GenerateAdminTechLogError(f"It is not possible to generate admintech log for {device_id}")
            time.sleep(polling_interval)
            polling_timer -= polling_interval
        raise GenerateAdminTechLogError(f'It is not possible to generate admintech log for {device_id}')

    def _get_token_id(self, filename: str) -> str:
        admin_techs = self.get_all()
        for admin_tech in admin_techs:
            if filename == admin_tech.filename:
                return admin_tech.token_id
        raise RequestTokenIdNotFound(
            f"requestTokenId of admin tech generation request not found for file name: {filename}"
        )

    def delete(self, filename: str) -> Response:
        """Deletes admin tech logs for a device.
        Args:
            filename: name of admin_tech file
        Returns:
            response: http response for delete operation
        """

        token_id = self._get_token_id(filename)
        response = self.session.delete(f"/dataservice/device/tools/admintech/{token_id}")
        if response.status_code == 200:
            logger.info(f"Deleted AdminTech file {filename} on remote")
        return response

    def download(self, filename: str, download_dir: Optional[Path] = None) -> Path:
        """Downloads admintech log for a device.
        Args:
            filename: name of admin_tech file
            download_dir: download directory (defaults to current working directory)
        Returns:
            path to downloaded admin_tech file
        """
        if not download_dir:
            download_dir = Path.cwd()
        download_path = download_dir / filename
        url = f"/dataservice/device/tools/admintech/download/{filename}"
        if self.session.get_file(url=url, filename=download_path).status_code != 200:
            raise DownloadAdminTechLogError(f"Cannot download admin tech file: {filename} from remote")
        logger.info(f"Downloaded AdminTech file to: {download_path}")
        return download_path



""" Methods for setting up packet capture session,
   and download .pcap session file.

    Returns:
        status: Status
"""
import logging
import time
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Iterator, Optional

from vmngclient.api.basic_api import DeviceStateAPI
from vmngclient.dataclasses import Device, PacketSetup, Status
from vmngclient.session import vManageSession
from vmngclient.utils.creation_tools import create_dataclass

logger = logging.getLogger(__name__)


class DownloadStatus(Enum):
    COMPLETED = "COMPLETED"
    IMPOSSIBLE = "IMPOSSIBLE"
    FILESIZE = None


class PacketCaptureAPI:
    def __init__(self, session: vManageSession, vpn: str = "0", interface: str = "ge0/1", status=None) -> None:
        self.session = session
        self.vpn = vpn
        self.interface = interface
        self.status = status

    def get_packets(self, device: Device, duration_time=20) -> Status:
        """Initate packet capture process.

        Args:
            device (Device): Device class object
            duration_time (int, optional): Duration od packet capturing . Defaults to 20.

        Returns:
            Status
        """
        with DeviceStateAPI(self.session).enable_data_stream():
            try:
                with self.channel(device):
                    with self.start_stop():
                        time.sleep(duration_time)
            except PermissionError as err:
                failed_status_dict = {
                    "fileDownloadStatus": DownloadStatus.IMPOSSIBLE.value,
                    "fileSize": DownloadStatus.FILESIZE.value,
                }
                self.status = create_dataclass(Status, failed_status_dict)
                logger.error(str(err))
            return self.status

    @contextmanager
    def channel(self, device: Device) -> Iterator:
        """Creates packet capture session.

        Args:
            device (Device): Device class object

        Raises:
            PermissionError: if already another packet capture session is active

        Yields:
            Iterator: packet_channel which contains session ID
        """
        query = {
            "deviceUUID": device.uuid,
            "vpn": self.vpn,
            "interface": self.interface,
            "type": "control",
        }

        try:
            url_path = r"/dataservice/stream/device/capture"
            packet_setup = self.session.post(url=url_path, json=query).json()  # TODO check
            self.packet_channel = create_dataclass(PacketSetup, packet_setup)
            if self.packet_channel.is_new_session is True:
                yield self.packet_channel
            else:
                self.status = None
                raise PermissionError("Can't start new session, another is already open")
        finally:
            for _ in range(3):
                time.sleep(10)
                self.status = self.get_status(self.packet_channel)
                if self.status.file_download_status == DownloadStatus.COMPLETED.value:
                    self.download_capture_session(self.packet_channel, device)
                    url_path = f"/dataservice/stream/device/capture/disable/{self.packet_channel.session_id}"
                    self.session.get_json(url_path)
                    break

    @contextmanager
    def start_stop(self) -> Iterator:
        """Start and stops packet capturing.

        Yields:
            Iterator: None
        """
        try:
            url_path = f"/dataservice/stream/device/capture/start/{self.packet_channel.session_id}"
            self.session.get_json(url_path)
            yield None

        finally:
            url_path = f"/dataservice/stream/device/capture/stop/{self.packet_channel.session_id}"
            self.session.get_json(url_path)

    def get_interface_name(self, device: Device) -> str:

        url_path = f"/dataservice/device/interface/synced?deviceId={device.local_system_ip}"
        ifname = dict(self.session.get_json(url_path))
        if_name = str(ifname["data"][0]["ifname"])
        return if_name

    def download_capture_session(self, packet: PacketSetup, device: Device, file_path: Optional[str] = None) -> bool:
        url_path = f"/dataservice/stream/device/capture/download/{packet.session_id}"
        full_url = self.session.get_full_url(url_path)
        download_packet = self.session.get_data(url=full_url)
        if file_path is None:
            file_path = f"{Path(__file__).parents[0]}/{device.uuid}.pcap"

        with open(file_path, "wb") as file:
            with download_packet as raw:
                file.write(raw.read())
        return True

    def get_status(self, packet_channel: PacketSetup) -> Status:
        url_path = f"/dataservice/stream/device/capture/status/{packet_channel.session_id}"
        self.status = dict(self.session.get_json(url_path))
        self.status = create_dataclass(Status, self.status)
        return self.status



import unittest
from unittest.mock import patch

from vmngclient.api.omp_api import OmpAPI
from vmngclient.dataclasses import (
    OmpAdvertisedRouteData,
    OmpAdvertisedTlocData,
    OmpPeerData,
    OmpReceivedRouteData,
    OmpReceivedTlocData,
    OmpServiceData,
    OmpSummaryData,
)
from vmngclient.utils.creation_tools import create_dataclass


class TestOmpAPI(unittest.TestCase):
    @patch('vmngclient.session.Session')
    def test_omp_peers(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.omp_peer
        # Act
        answer = OmpAPI(mock_session).get_omp_peers(self.device_id)
        # Assert
        self.assertEqual(answer, self.omp_peer_dataclass)

    @patch('vmngclient.session.Session')
    def test_omp_peers_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = OmpAPI(mock_session).get_omp_peers(self.device_id)
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.Session')
    def test_advertised_routes(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.advertised_routes
        # Act
        answer = OmpAPI(mock_session).get_advertised_routes(self.device_id)
        # Assert
        self.assertEqual(answer, self.advertised_routes_dataclass)

    @patch('vmngclient.session.Session')
    def test_advertised_routes_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = OmpAPI(mock_session).get_advertised_routes(self.device_id)
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.Session')
    def test_received_routes(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.received_routes
        # Act
        answer = OmpAPI(mock_session).get_received_routes(self.device_id)
        # Assert
        self.assertEqual(answer, self.received_routes_dataclass)

    @patch('vmngclient.session.Session')
    def test_received_routes_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = OmpAPI(mock_session).get_received_routes(self.device_id)
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.Session')
    def test_advertised_tlocs(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.advertised_tlocs
        # Act
        answer = OmpAPI(mock_session).get_advertised_tlocs(self.device_id)
        # Assert
        self.assertEqual(answer, self.advertised_tlocs_dataclass)

    @patch('vmngclient.session.Session')
    def test_advertised_tlocs_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = OmpAPI(mock_session).get_advertised_tlocs(self.device_id)
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.Session')
    def test_received_tlocs(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.received_tlocs
        # Act
        answer = OmpAPI(mock_session).get_received_tlocs(self.device_id)
        # Assert
        self.assertEqual(answer, self.received_tlocs_dataclass)

    @patch('vmngclient.session.Session')
    def test_received_tlocs_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = OmpAPI(mock_session).get_received_tlocs(self.device_id)
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.Session')
    def test_services(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.services
        # Act
        answer = OmpAPI(mock_session).get_services(self.device_id)
        # Assert
        self.assertEqual(answer, self.services_dataclass)

    @patch('vmngclient.session.Session')
    def test_services_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = OmpAPI(mock_session).get_services(self.device_id)
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.Session')
    def test_omp_summary(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.omp_summary
        # Act
        answer = OmpAPI(mock_session).get_omp_summary(self.device_id)
        # Assert
        self.assertEqual(answer, self.omp_summary_dataclass)

    @patch('vmngclient.session.Session')
    def test_omp_summary_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = OmpAPI(mock_session).get_omp_summary(self.device_id)
        # Assert
        self.assertEqual(answer, [])

    def setUp(self) -> None:
        self.device_id = "1.1.1.1"
        self.omp_peer = [
            {
                "domain-id": 1,
                "vdevice-name": "169.254.10.10",
                "refresh": "supported",
                "site-id": 129,
                "type": "vsmart",
                "vdevice-host-name": "vm3",
                "up-time-date": 1643135160000,
                "vdevice-dataKey": "169.254.10.10-172.16.253.129-vsmart",
                "peer": "172.16.253.129",
                "up-time": "0:12:57:30",
                "legit": "yes",
                "lastupdated": 1643181828115,
                "state": "up",
            },
            {
                "domain-id": 1,
                "vdevice-name": "169.254.10.10",
                "refresh": "supported",
                "site-id": 130,
                "type": "vsmart",
                "vdevice-host-name": "vm3",
                "up-time-date": 1643135160000,
                "vdevice-dataKey": "169.254.10.10-172.16.255.130-vsmart",
                "peer": "172.16.255.130",
                "up-time": "0:12:57:30",
                "legit": "yes",
                "lastupdated": 1643181828115,
                "state": "up",
            },
        ]
        self.omp_peer_dataclass = [create_dataclass(OmpPeerData, item) for item in self.omp_peer]
        self.advertised_routes = [
            {
                "overlay-id": "1",
                "color": "default",
                "vdevice-name": "169.254.10.10",
                "prefix": "100.100.3.0/24",
                "ip": "172.16.254.2",
                "label": "1069",
                "encap": "ipsec",
                "site-id": "3",
                "originator": "172.16.254.2",
                "vpn-id": "10",
                "vdevice-host-name": "vm3",
                "path-id": "65",
                "protocol": "connected",
                "vdevice-dataKey": "169.254.10.10-10",
                "metric": "0",
                "lastupdated": 1643179668956,
                "to-peer": "172.16.253.129",
            },
            {
                "overlay-id": "1",
                "color": "default",
                "vdevice-name": "169.254.10.10",
                "prefix": "100.100.3.0/24",
                "ip": "172.16.254.2",
                "label": "1069",
                "encap": "ipsec",
                "site-id": "3",
                "originator": "172.16.254.2",
                "vpn-id": "10",
                "vdevice-host-name": "vm3",
                "path-id": "65",
                "protocol": "connected",
                "vdevice-dataKey": "169.254.10.10-10",
                "metric": "0",
                "lastupdated": 1643179668956,
                "to-peer": "172.16.255.130",
            },
        ]
        self.advertised_routes_dataclass = [
            create_dataclass(OmpAdvertisedRouteData, item) for item in self.advertised_routes
        ]
        self.received_routes = [
            {
                "overlay-id": "1",
                "color": "default",
                "vdevice-name": "169.254.10.10",
                "prefix": "10.0.5.0/24",
                "ip": "172.16.254.4",
                "from-peer": "172.16.253.129",
                "label": "1047",
                "encap": "ipsec",
                "site-id": "5",
                "originator": "172.16.254.4",
                "vpn-id": "10",
                "vdevice-host-name": "vm3",
                "path-id": "42",
                "protocol": "connected",
                "vdevice-dataKey": "169.254.10.10-10",
                "metric": "0",
                "lastupdated": 1643182041519,
                "attribute-type": "installed",
                "status": "C I R",
            },
            {
                "overlay-id": "1",
                "color": "default",
                "vdevice-name": "169.254.10.10",
                "prefix": "100.100.3.0/24",
                "ip": "172.16.254.2",
                "from-peer": "0.0.0.0",
                "label": "1069",
                "encap": "ipsec",
                "site-id": "3",
                "originator": "172.16.254.2",
                "vpn-id": "10",
                "vdevice-host-name": "vm3",
                "path-id": "65",
                "protocol": "connected",
                "vdevice-dataKey": "169.254.10.10-10",
                "metric": "0",
                "lastupdated": 1643182041520,
                "attribute-type": "installed",
                "status": "C Red R",
            },
        ]
        self.received_routes_dataclass = [create_dataclass(OmpReceivedRouteData, item) for item in self.received_routes]
        self.advertised_tlocs = [
            {
                "color": "default",
                "vdevice-name": "169.254.10.10",
                "ip": "172.16.254.2",
                "tloc-auth-type": "sha1-hmac ah-sha1-hmac",
                "preference": "0",
                "weight": "1",
                "encap": "ipsec",
                "site-id": "3",
                "originator": "172.16.254.2",
                "vdevice-host-name": "vm3",
                "tloc-public-ip": "10.101.3.3",
                "tloc-public-port": "12406",
                "tloc-private-ip": "10.101.3.3",
                "vdevice-dataKey": "169.254.10.10-ipv4-172.16.254.2",
                "tloc-private-port": "12406",
                "tloc-spi": "286",
                "lastupdated": 1643179953349,
                "tloc-encrypt-type": "aes256",
                "tloc-proto": "0",
                "address-family": "ipv4",
                "to-peer": "172.16.253.129",
            },
            {
                "color": "default",
                "vdevice-name": "169.254.10.10",
                "ip": "172.16.254.2",
                "tloc-auth-type": "sha1-hmac ah-sha1-hmac",
                "preference": "0",
                "weight": "1",
                "encap": "ipsec",
                "site-id": "3",
                "originator": "172.16.254.2",
                "vdevice-host-name": "vm3",
                "tloc-public-ip": "10.101.3.3",
                "tloc-public-port": "12406",
                "tloc-private-ip": "10.101.3.3",
                "vdevice-dataKey": "169.254.10.10-ipv4-172.16.254.2",
                "tloc-private-port": "12406",
                "tloc-spi": "286",
                "lastupdated": 1643179953351,
                "tloc-encrypt-type": "aes256",
                "tloc-proto": "0",
                "address-family": "ipv4",
                "to-peer": "172.16.255.130",
            },
        ]
        self.advertised_tlocs_dataclass = [
            create_dataclass(OmpAdvertisedTlocData, item) for item in self.advertised_tlocs
        ]
        self.received_tlocs = [
            {
                "bfd-status": "up",
                "color": "default",
                "vdevice-name": "169.254.10.10",
                "ip": "172.16.254.1",
                "tloc-auth-type": "sha1-hmac ah-sha1-hmac",
                "preference": "0",
                "from-peer": "172.16.253.129",
                "weight": "1",
                "encap": "ipsec",
                "site-id": "2",
                "originator": "172.16.254.1",
                "vdevice-host-name": "vm3",
                "tloc-public-ip": "10.101.2.2",
                "tloc-public-port": "12386",
                "tloc-private-ip": "10.101.2.2",
                "vdevice-dataKey": "169.254.10.10-ipv4-172.16.254.1",
                "tloc-private-port": "12386",
                "tloc-spi": "263",
                "lastupdated": 1643185163270,
                "tloc-encrypt-type": "aes256",
                "tloc-proto": "0",
                "address-family": "ipv4",
            },
            {
                "bfd-status": "up",
                "color": "default",
                "vdevice-name": "169.254.10.10",
                "ip": "172.16.254.4",
                "tloc-auth-type": "sha1-hmac ah-sha1-hmac",
                "preference": "0",
                "from-peer": "172.16.255.130",
                "weight": "1",
                "encap": "ipsec",
                "site-id": "5",
                "originator": "172.16.254.4",
                "vdevice-host-name": "vm3",
                "tloc-public-ip": "10.102.5.5",
                "tloc-public-port": "12406",
                "tloc-private-ip": "10.102.5.5",
                "vdevice-dataKey": "169.254.10.10-ipv4-172.16.254.4",
                "tloc-private-port": "12406",
                "tloc-spi": "288",
                "lastupdated": 1643185163286,
                "tloc-encrypt-type": "aes256",
                "tloc-proto": "0",
                "address-family": "ipv4",
            },
        ]
        self.received_tlocs_dataclass = [create_dataclass(OmpReceivedTlocData, item) for item in self.received_tlocs]
        self.services = [
            {
                "path-id": "65",
                "vdevice-dataKey": "169.254.10.10-10-VPN-172.16.254.2",
                "vdevice-name": "169.254.10.10",
                "service": "VPN",
                "from-peer": "0.0.0.0",
                "lastupdated": 1643250832867,
                "originator": "172.16.254.2",
                "label": "1077",
                "vpn-id": "10",
                "vdevice-host-name": "vm3",
                "address-family": "ipv4",
                "status": "C Red R",
            },
            {
                "vdevice-dataKey": "169.254.10.10-10-VPN-172.16.254.2",
                "vdevice-name": "169.254.10.10",
                "service": "VPN",
                "lastupdated": 1643250832869,
                "originator": "172.16.254.2",
                "vpn-id": "10",
                "vdevice-host-name": "vm3",
                "address-family": "ipv4",
                "to-peer": "172.16.255.130",
            },
        ]
        self.services_dataclass = [create_dataclass(OmpServiceData, item) for item in self.services]
        self.omp_summary = [
            {
                "tlocs-sent": 4,
                "policy-sent": 0,
                "mcast-routes-sent": 0,
                "packets-sent": 81579,
                "vdevice-name": "169.254.10.10",
                "inform-sent": 14,
                "packets-received": 81956,
                "routes-received": 18,
                "tlocs-received": 22,
                "mcast-routes-installed": 0,
                "update-sent": 84,
                "devicetype": "vedge",
                "mcast-routes-received": 0,
                "hello-received": 81472,
                "alert-sent": 0,
                "update-received": 468,
                "vsmart-peers": 2,
                "operstate": "UP",
                "policy-received": 0,
                "handshake-received": 2,
                "handshake-sent": 2,
                "alert-received": 0,
                "services-sent": 4,
                "vdevice-host-name": "vm3",
                "inform-received": 14,
                "vdevice-dataKey": "169.254.10.10--",
                "tlocs-installed": 10,
                "services-installed": 0,
                "ompuptime": "9:11:16:34",
                "services-received": 2,
                "lastupdated": 1645494228477,
                "routes-sent": 4,
                "hello-sent": 81479,
                "routes-installed": 8,
                "adminstate": "UP",
            }
        ]
        self.omp_summary_dataclass = [create_dataclass(OmpSummaryData, item) for item in self.omp_summary]



from enum import Enum


class DeviceModel(Enum):

    # controlers
    VSMART = "vsmart"
    VBOND = "vedge-cloud"
    VMANAGE = "vmanage"

    # vedges
    VEDGE_ISR1100_6G = "vedge-ISR1100-6G"
    VEDGE_ISR1100X_6G = "vedge-ISR1100X-6G"
    VEDGE_ISR1100_4G = "vedge-ISR1100-4G"
    VEDGE_ISR1100X_4G = "vedge-ISR1100X-4G"
    VEDGE_ISR1100_4GLTE = "vedge-ISR1100-4GLTE"
    VEDGE = "vedge-cloud"
    VEDGE_1000 = "vedge-1000"
    VEDGE_2000 = "vedge-2000"
    VEDGE_100 = "vedge-100"
    VEDGE_100_B = "vedge-100-B"
    VEDGE_100_WM = "vedge-100-WM"
    VEDGE_100_M = "vedge-100-M"
    VEDGE_5000 = "vedge-5000"

    # IR1101
    VEDGE_IR_1101 = "vedge-IR-1101"
    VEDGE_ESR_6300 = "vedge-ESR-6300"
    VEDGE_IR_1821 = "vedge-IR-1821"
    VEDGE_IR_1831 = "vedge-IR-1831"
    VEDGE_IR_1833 = "vedge-IR-1833"
    VEDGE_IR_1835 = "vedge-IR-1835"

    # ASR devices
    VEDGE_ASR_1001_X = "vedge-ASR-1001-X"
    VEDGE_ASR_1002_X = "vedge-ASR-1002-X"
    VEDGE_ASR_1002_HX = "vedge-ASR-1002-HX"
    VEDGE_ASR_1001_HX = "vedge-ASR-1001-HX"
    VEDGE_C8500L_8G4X = "vedge-C8500L-8G4X"
    VEDGE_C8500_12X4QC = "vedge-C8500-12X4QC"
    VEDGE_C8500_12X = "vedge-C8500-12X"
    VEDGE_C8500L_8S4X = "vedge-C8500L-8S4X"
    VEDGE_ASR_1006_X = "vedge-ASR-1006-X"
    VEDGE_C8500_20X6C = "vedge-C8500-20X6C"

    # CSR
    VEDGE_CSR_1000V = "vedge-CSR-1000v"
    VEDGE_C8000V = "vedge-C8000V"

    # ISR
    VEDGE_ISR_4331 = "vedge-ISR-4331"
    VEDGE_ISR_4431 = "vedge-ISR-4431"
    VEDGE_ISR_4461 = "vedge-ISR-4461"
    VEDGE_ISR_4451_X = "vedge-ISR-4451-X"
    VEDGE_ISR_4321 = "vedge-ISR-4321"
    VEDGE_ISR_4351 = "vedge-ISR-4351"
    VEDGE_ISR_4221 = "vedge-ISR-4221"
    VEDGE_ISR_4221X = "vedge-ISR-4221X"

    # C1100 DSL
    VEDGE_C1111_8PW = "vedge-C1111-8PW"
    VEDGE_C1111_8PLTELAW = "vedge-C1111-8PLTELAW"
    VEDGE_C1111_8PLTEEAW = "vedge-C1111-8PLTEEAW"
    VEDGE_C1113_8PMLTEEA = "vedge-C1113-8PMLTEEA"
    VEDGE_C1116_4P = "vedge-C1116-4P"
    VEDGE_C1116_4PLTEEA = "vedge-C1116-4PLTEEA"
    VEDGE_C1117_4P = "vedge-C1117-4P"
    VEDGE_C1117_4PM = "vedge-C1117-4PM"
    VEDGE_C1117_4PLTEEA = "vedge-C1117-4PLTEEA"
    VEDGE_C1111_8PLTELA = "vedge-C1111-8PLTELA"
    VEDGE_C1111_8PLTEEA = "vedge-C1111-8PLTEEA"

    # ISRv
    VEDGE_ISRV = "vedge-ISRv"



import unittest
from unittest.mock import MagicMock, Mock, patch

from ciscoconfparse import CiscoConfParse  # type: ignore

from vmngclient.api.templates import AttachedError, NameAlreadyExistError, NotFoundError, TemplateAPI
from vmngclient.dataclasses import Device, Template
from vmngclient.utils.creation_tools import create_dataclass
from vmngclient.utils.operation_status import OperationStatus
from vmngclient.utils.personality import Personality
from vmngclient.utils.reachability import Reachability


class TestTemplateAPI(unittest.TestCase):
    def setUp(self):
        self.data_template = [
            {
                'deviceType': 'vedge-C8000V',
                'lastUpdatedBy': 'user',
                'resourceGroup': 'global',
                'templateClass': 'cedge',
                'configType': 'file',
                'templateId': 'dummy_id_1',
                'factoryDefault': False,
                'templateName': 'template_1',
                'devicesAttached': 0,
                'templateDescription': 'dummy template 1',
                'draftMode': 'Disabled',
                'lastUpdatedOn': 0,
                'templateAttached': 0,
            },
            {
                'deviceType': 'vedge-cloud',
                'lastUpdatedBy': 'admin',
                'resourceGroup': 'global',
                'templateClass': 'vedge',
                'configType': 'file',
                'templateId': 'dummy_id_2',
                'factoryDefault': False,
                'templateName': 'template_2',
                'devicesAttached': 1,
                'templateDescription': 'dummy template 2',
                'draftMode': 'Disabled',
                'lastUpdatedOn': 0,
                'templateAttached': 0,
            },
        ]
        self.templates = [create_dataclass(Template, template) for template in self.data_template]
        self.device_info = Device(
            personality=Personality.EDGE,
            uuid='dummy_uuid',
            id='162.168.0.1',
            hostname='dummy_host',
            reachability=Reachability.REACHABLE,
            local_system_ip='192.168.0.1',
            memUsage=1.0,
            connected_vManages=['192.168.0.1'],
            model='vedge-cloud',
            status='normal',
        )

    @patch('vmngclient.session.vManageSession')
    def test_templates_success(self, mock_session):

        # Arrange
        mock_session.get_data.return_value = self.data_template

        # Act
        answer = TemplateAPI(mock_session).templates

        # Assert
        self.assertEqual(answer, self.templates)

    @patch('vmngclient.session.vManageSession')
    def test_templates_no_data(self, mock_session):

        # Arrange
        mock_session.get_data.return_value = []

        # Act
        answer = TemplateAPI(mock_session).templates

        # Assert
        self.assertEqual(answer, [])

    @patch.object(TemplateAPI, 'templates')
    def test_get_success(self, mock_templates):

        # Arrage
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.templates = self.templates

        # Act
        answer = [test_object.get(template.name) for template in self.templates]

        # Assert
        self.assertEqual(answer, self.templates)

    @patch.object(TemplateAPI, 'templates')
    def test_get_exception(self, mock_templates):

        # Arrage
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.templates = self.templates

        # Act
        def answer():
            return test_object.get('no_exist_name')

        # Assert
        self.assertRaises(NotFoundError, answer)

    @patch.object(TemplateAPI, 'templates')
    def test_get_id_success(self, mock_templates):

        # Arrage
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.templates = self.templates

        # Act
        answer = [test_object.get_id(template.name) for template in self.templates]

        ids = [template.id for template in self.templates]
        # Assert
        self.assertEqual(answer, ids)

    @patch.object(TemplateAPI, 'templates')
    def test_get_id_exception(self, mock_templates):

        # Arrage
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.templates = self.templates

        # Act
        def answer():
            return test_object.get_id('no_exist_name')

        # Assert
        self.assertRaises(NotFoundError, answer)

    @patch.object(TemplateAPI, "get_operation_status")
    def test_wait_for_complete_true(self, mock_operation):

        # Arrage
        MockOperatrion = Mock()
        mock_operation.return_value = MockOperatrion
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.get_operation_status.return_value = [OperationStatus.SUCCESS, OperationStatus.SUCCESS]
        operation_id = 'operation_id'

        # Act
        answer = test_object.wait_for_complete(operation_id)

        # Assert
        self.assertTrue(answer)

    @patch.object(TemplateAPI, "get_operation_status")
    def test_wait_for_complete_false(self, mock_operation):

        # Arrage
        MockOperatrion = Mock()
        mock_operation.return_value = MockOperatrion
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.get_operation_status.return_value = [OperationStatus.IN_PROGRESS, OperationStatus.SUCCESS]
        operation_id = 'operation_id'

        # Act
        answer = test_object.wait_for_complete(operation_id, timeout_seconds=2, sleep_seconds=1)

        # Assert
        self.assertFalse(answer)

    @patch('vmngclient.session.vManageSession')
    @patch.object(TemplateAPI, 'templates')
    @patch.object(TemplateAPI, "wait_for_complete")
    def test_attach_exist_template(self, mock_session, mock_wait_for_complete, mock_templates):

        # Arrage
        mock_session.post_json.return_value = {"id": "operation_id"}
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # mock wait complete
        MockWaitComplete = Mock()
        mock_wait_for_complete.return_value = MockWaitComplete
        test_object.wait_for_complete.return_value = True

        # Act
        answer = test_object.attach('template_1', self.device_info)

        # Assert
        self.assertTrue(answer)

    @patch('vmngclient.session.vManageSession')
    @patch.object(TemplateAPI, 'templates')
    @patch.object(TemplateAPI, "wait_for_complete")
    def test_attach_no_exist_template(self, mock_session, mock_wait_for_complete, mock_templates):

        # Arrage
        mock_session.post_json.return_value = {"id": "operation_id"}
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # mock wait complete
        MockWaitComplete = Mock()
        mock_wait_for_complete.return_value = MockWaitComplete
        test_object.wait_for_complete.return_value = True

        # Act
        answer = test_object.attach('no_exist_template', self.device_info)

        # Assert
        self.assertFalse(answer)

    @patch('vmngclient.session.vManageSession')
    @patch.object(TemplateAPI, "wait_for_complete")
    def test_device_to_cli_true(self, mock_session, mock_wait_for_complete):

        # Arrage
        mock_session.post_json.return_value = {"id": "operation_id"}
        test_object = TemplateAPI(mock_session)

        # mock wait complete
        MockWaitComplete = Mock()
        mock_wait_for_complete.return_value = MockWaitComplete
        test_object.wait_for_complete.return_value = True

        # Act
        answer = test_object.device_to_cli(self.device_info)

        # Assert
        self.assertTrue(answer)

    @patch('vmngclient.session.vManageSession')
    def test_get_operatrion_status_success(self, mock_session):

        # Arrage
        mock_session.get_data.return_value = [
            {"status": "In progress"},
            {"status": "Failure"},
            {"status": "Success"},
            {"status": "Scheduled"},
        ]
        test_object = TemplateAPI(mock_session)
        operation_id = 'dummy_id'

        # Act
        answer = test_object.get_operation_status(operation_id)

        # Assert
        self.assertEqual(
            answer,
            [OperationStatus.IN_PROGRESS, OperationStatus.FAILURE, OperationStatus.SUCCESS, OperationStatus.SCHEDULED],
        )

    @patch('vmngclient.session.vManageSession')
    def test_get_operatrion_status_no_data(self, mock_session):

        # Arrage
        mock_session.get_data.return_value = []
        test_object = TemplateAPI(mock_session)
        operation_id = 'dummy_id'

        # Act
        answer = test_object.get_operation_status(operation_id)

        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.vManageSession')
    @patch.object(TemplateAPI, 'templates')
    def test_delete_success(self, mock_session, mock_templates):

        # Arrage
        MockResponse = MagicMock()
        MockResponse.status_code = 200
        mock_session.delete.return_value = MockResponse
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # Act
        answer = test_object.delete('template_1')
        # Assert
        self.assertTrue(answer)

    @patch('vmngclient.session.vManageSession')
    @patch.object(TemplateAPI, 'templates')
    def test_delete_wrong_status(self, mock_session, mock_templates):

        # Arrage
        MockResponse = MagicMock()
        MockResponse.ok = False
        mock_session.delete.return_value = MockResponse
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # Act
        answer = test_object.delete('template_1')

        # Assert
        self.assertFalse(answer)

    @patch('vmngclient.session.vManageSession')
    @patch.object(TemplateAPI, 'templates')
    def test_delete_exception(self, mock_session, mock_templates):

        # Arrage
        MockResponse = MagicMock()
        MockResponse.status = 404
        mock_session.delete.return_value = MockResponse
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # Act
        def answer():
            return test_object.delete('template_2')

        # Assert
        self.assertRaises(AttachedError, answer)

    @patch('vmngclient.session.vManageSession')
    @patch.object(TemplateAPI, 'templates')
    def test_create_exception(self, mock_session, mock_templates):

        # Arrage
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates
        config = CiscoConfParse([])

        # Act
        def answer():
            return test_object.create(self.device_info, 'template_1', 'new_description', config)

        # Assert
        self.assertRaises(NameAlreadyExistError, answer)



from unittest import TestCase
from unittest.mock import Mock, patch

from parameterized import parameterized  # type: ignore
from tenacity import RetryError

from vmngclient.api.basic_api import DeviceField, DevicesAPI, DeviceStateAPI, FailedSend
from vmngclient.dataclasses import BfdSessionData, Connection, Device, Reboot, WanInterface
from vmngclient.utils.creation_tools import create_dataclass
from vmngclient.utils.personality import Personality


class TestDevicesAPI(TestCase):
    def setUp(self) -> None:
        self.devices = [
            {  # vmanage
                "device-model": "vmanage",
                "deviceId": "1.1.1.1",
                "uuid": "aaaaaaaa-6169-445c-8e49-c0bdaaaaaaa",
                "cpuLoad": 4.71,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.1",
                "board-serial": "11122233",
                "personality": "vmanage",
                "memUsage": 57.0,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm200",
                "cpuState": "normal",
                "chassis-number": "aaaaaaaa-6aa9-445c-8e49-c0aaaaaaaaa9",
            },
            {  # vsmart
                "device-model": "vsmart",
                "deviceId": "1.1.1.3",
                "uuid": "bbcccccc-6169-445c-8e49-c0bdccccccc",
                "cpuLoad": 2.76,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.3",
                "board-serial": "11223399",
                "personality": "vsmart",
                "memUsage": 28.0,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm129",
                "cpuState": "normal",
                "chassis-number": "abbbbbbb-6169-445c-8e49-c0bccccccccc",
            },
            {  # vbond
                "device-model": "vedge-cloud",
                "deviceId": "1.1.1.2",
                "uuid": "bbbbbbbb-6169-445c-8e49-c0bdaaaaaaa",
                "cpuLoad": 1.76,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.2",
                "board-serial": "11223344",
                "personality": "vbond",
                "memUsage": 26.0,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm128",
                "cpuState": "normal",
                "chassis-number": "abbbbbbb-6169-445c-8e49-c0bbbbbbbbb9",
            },
            {  # vedge
                "device-model": "vedge-cloud",
                "deviceId": "169.254.10.10",
                "uuid": "bc2a78ac-a06e-40fd-b2b7-1b1e062f3f9e",
                "cpuLoad": 1.46,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "172.16.254.2",
                "board-serial": "12345708",
                "personality": "vedge",
                "memUsage": 24.1,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm1",
                "cpuState": "normal",
                "chassis-number": "bc2a78ac-a06e-40fd-b2b7-1b1e062f3f9e",
            },
        ]
        self.devices_dataclass = [create_dataclass(Device, device) for device in self.devices]
        self.controllers_dataclass = [create_dataclass(Device, device) for device in self.devices[:2]]
        self.orchestrators_dataclass = [create_dataclass(Device, self.devices[2])]
        self.edges_dataclass = [create_dataclass(Device, self.devices[3])]
        self.vsmarts_dataclass = [create_dataclass(Device, self.devices[1])]
        self.system_ips_list = [device["local-system-ip"] for device in self.devices]
        self.ips_list = [device["deviceId"] for device in self.devices]
        self.tenants = [
            {
                'flakeId': 11111,
                'orgName': 'my-org Inc',
                'samlSpInfo': '',
                'subDomain': 'sub1.domain.com',
                'vBondAddress': '1.1.1.1',
                'oldIdpMetadata': '',
                'configDBClusterServiceName': '',
                'vSmarts': ['aaaaaaaa-aaac-42f4-b2c7-1aaaaaaaaaaaa', 'bbbbbbbb-9d8a-453c-841f-4bbbbbbbbbbb'],
                'mode': 'off',
                'idpMetadata': '',
                'createdAt': 1111111111111,
                '@rid': 1111,
                'tenantId': 'aaaaaaaaaa-9ca8-42fd-8290-67aaaaaaaaaa',
                'name': 'sub1',
                'wanEdgeForecast': '100',
                'spMetadata': '',
                'state': 'READY',
                'wanEdgePresent': 1,
                'desc': 'This is sub1',
            },
            {
                'flakeId': 22222,
                'orgName': 'my-org Inc',
                'samlSpInfo': '',
                'subDomain': 'sub2.domain.com',
                'vBondAddress': '1.1.1.2',
                'oldIdpMetadata': '',
                'configDBClusterServiceName': '',
                'vSmarts': ['ccccccccccccccc-42f4-b2c7-1cccccccccc', 'ddddddddd-9d8a-453c-841f-4dddddddddd'],
                'mode': 'off',
                'idpMetadata': '',
                'createdAt': 2222222222222,
                '@rid': 2222,
                'tenantId': 'ccccccccca-9ca8-42fd-8290-67aaaccccccc',
                'name': 'sub2',
                'wanEdgeForecast': '100',
                'spMetadata': '',
                'state': 'READY',
                'wanEdgePresent': 1,
                'desc': 'This is sub2',
            },
        ]
        self.devices_vbonds = [self.devices[2]]
        self.devices_vsmatrs = [self.devices[1]]
        self.devices_edges = [self.devices[3]]

    @patch.object(DevicesAPI, 'devices')
    def test_controllers(self, mock_devices):
        # Arrange
        MockDevices = Mock()
        mock_devices.return_value = MockDevices
        session = Mock()
        test_object = DevicesAPI(session)
        test_object.devices = self.devices_dataclass
        # Act
        answer = test_object.controllers
        # Assert
        self.assertEqual(answer, self.controllers_dataclass)

    @patch.object(DevicesAPI, 'devices')
    def test_orchestrators(self, mock_devices):
        # Arrange
        MockDevices = Mock()
        mock_devices.return_value = MockDevices
        session = Mock()
        test_object = DevicesAPI(session)
        test_object.devices = self.devices_dataclass
        # Act
        answer = test_object.orchestrators
        # Assert
        self.assertEqual(answer, self.orchestrators_dataclass)

    @patch.object(DevicesAPI, 'devices')
    def test_edges(self, mock_devices):
        # Arrange
        MockDevices = Mock()
        mock_devices.return_value = MockDevices
        session = Mock()
        test_object = DevicesAPI(session)
        test_object.devices = self.devices_dataclass
        # Act
        answer = test_object.edges
        # Assert
        self.assertEqual(answer, self.edges_dataclass)

    @patch.object(DevicesAPI, 'devices')
    def test_vsmarts(self, mock_vsmarts):
        # Arrange
        MockDevices = Mock()
        mock_vsmarts.return_value = MockDevices
        session = Mock()
        test_object = DevicesAPI(session)
        test_object.devices = self.devices_dataclass
        # Act
        answer = test_object.vsmarts
        # Assert
        self.assertEqual(answer, self.vsmarts_dataclass)

    @patch.object(DevicesAPI, 'devices')
    def test_system_ips(self, mock_devices):
        # Arrange
        MockDevices = Mock()
        mock_devices.return_value = MockDevices
        session = Mock()
        test_object = DevicesAPI(session)
        test_object.devices = self.devices_dataclass
        # Act
        answer = test_object.system_ips
        # Assert
        self.assertEqual(answer, self.system_ips_list)

    @patch.object(DevicesAPI, 'devices')
    def test_ips(self, mock_devices):
        # Arrange
        MockDevices = Mock()
        mock_devices.return_value = MockDevices
        session = Mock()
        test_object = DevicesAPI(session)
        test_object.devices = self.devices_dataclass
        # Act
        answer = test_object.ips
        # Assert
        self.assertEqual(answer, self.ips_list)

    @patch('vmngclient.session.vManageSession')
    def test_devices(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.devices
        # Act
        answer = DevicesAPI(mock_session).devices
        # Assert
        self.assertEqual(answer, self.devices_dataclass)

    def test_get_device_details(self):
        pass  # TODO fix method before test

    def test_count_devices(self):
        pass  # TODO fix method before test

    @patch('vmngclient.session.vManageSession')
    def test_get_tenants(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.tenants
        # Act
        answer = DevicesAPI(mock_session).get_tenants()
        # Assert
        self.assertEqual(answer, self.tenants)

    @patch('vmngclient.session.vManageSession')
    def test_get_reachable_devices_for_vsmatrs(self, mock_session, personality=Personality.VSMART):
        # Arrange
        mock_session.get_data.return_value = [
            device for device in self.devices if device["personality"] == personality.value
        ]
        # Act
        answer = DevicesAPI(mock_session).get_reachable_devices(personality)
        # Assert
        self.assertEqual(answer, self.devices_vsmatrs)

    @patch('vmngclient.session.vManageSession')
    def test_get_reachable_devices_for_vbonds(self, mock_session, personality=Personality.VBOND):
        # Arrange
        mock_session.get_data.return_value = [
            device for device in self.devices if device["personality"] == personality.value
        ]
        # Act
        answer = DevicesAPI(mock_session).get_reachable_devices(personality)
        # Assert
        self.assertEqual(answer, self.devices_vbonds)

    @patch('vmngclient.session.vManageSession')
    def test_get_reachable_devices_for_edges(self, mock_session, personality=Personality.EDGE):
        # Arrange
        mock_session.get_data.return_value = [
            device for device in self.devices if device["personality"] == personality.value
        ]
        # Act
        answer = DevicesAPI(mock_session).get_reachable_devices(personality)
        # Assert
        self.assertEqual(answer, self.devices_edges)

    @patch('vmngclient.session.vManageSession')
    def test_get_reachable_devices_for_vmanage(self, mock_session, personality=Personality.VMANAGE):
        # Arrange
        mock_session.get_data.return_value = [
            device for device in self.devices if device["personality"] == personality.value
        ]

        # Act
        def answer():
            return DevicesAPI(mock_session).get_reachable_devices(personality)

        # Assert
        self.assertRaises(AssertionError, answer)

    @patch('vmngclient.session.vManageSession')
    def test_send_certificate_state_to_controllers(self, mock_session):
        # Arrange
        mock_session.post().json.return_value = {'id': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'}

        # Act
        answer = DevicesAPI(mock_session).send_certificate_state_to_controllers()

        # Assert
        self.assertTrue(answer)

    @patch('vmngclient.session.vManageSession')
    def test_send_certificate_state_to_controllers_error(self, mock_session):
        # Arrange
        mock_session.post().json.return_value = {}

        # Act
        def answer():
            return DevicesAPI(mock_session).send_certificate_state_to_controllers()

        # Assert
        self.assertRaises(FailedSend, answer)

    @parameterized.expand([["vm200", 0], ["vm129", 1], ["vm128", 2], ["vm1", 3]])
    @patch.object(DevicesAPI, 'devices')
    def test_get_by_hostname(self, hostname, device_number, mock_devices):
        # Arrange
        MockDevices = Mock()
        mock_devices.return_value = MockDevices
        session = Mock()
        test_object = DevicesAPI(session)
        test_object.devices = self.devices_dataclass
        # Act
        answer = test_object.get(DeviceField.HOSTNAME, hostname)
        # Assert
        self.assertEqual(answer, self.devices_dataclass[device_number])

    @parameterized.expand([["1.1.1.1", 0], ["1.1.1.3", 1], ["1.1.1.2", 2], ["169.254.10.10", 3]])
    @patch.object(DevicesAPI, 'devices')
    def test_get_by_id(self, device_id, device_number, mock_devices):
        # Arrange
        MockDevices = Mock()
        mock_devices.return_value = MockDevices
        session = Mock()
        test_object = DevicesAPI(session)
        test_object.devices = self.devices_dataclass
        # Act
        answer = test_object.get(DeviceField.ID, device_id)
        # Assert
        self.assertEqual(answer, self.devices_dataclass[device_number])


class TestDevicesStateAPI(TestCase):
    def setUp(self) -> None:
        self.crash_info = [
            {
                "core-time": "Tue Feb 15 04:14:38 UTC 2022",
                "vdevice-dataKey": "169.254.10.12-0",
                "vdevice-name": "169.254.10.12",
                "index": 0,
                "lastupdated": 1645064726542,
                "core-filename": "vm5_htx_19219_20220215-041430-UTC.core.gz",
                "core-time-date": 1644898478000,
                "vdevice-host-name": "vm5",
            },
            {
                "core-time": "Tue Feb 15 04:15:24 UTC 2022",
                "vdevice-dataKey": "169.254.10.12-1",
                "vdevice-name": "169.254.10.12",
                "index": 1,
                "lastupdated": 1645064726542,
                "core-filename": "vm5_htx_19539_20220215-041515-UTC.core.gz",
                "core-time-date": 1644898524000,
                "vdevice-host-name": "vm5",
            },
        ]
        self.connections_info = [
            {
                'system-ip': '1.1.1.1',
                'peer-type': 'vsmart',
                'state': 'up',
            },
            {
                'system-ip': '1.1.1.2',
                'peer-type': 'vedge',
                'state': 'up',
            },
        ]
        self.connections_info_dataclass = [create_dataclass(Connection, item) for item in self.connections_info]
        self.reboot_history = [
            {
                "reboot_date_time-date": 1642651714000,
                "lastupdated": 1642656685355,
                "vdevice-dataKey": "172.16.255.11-2022-01-20T04:08:34+00:00",
                "reboot_date_time": "2022-01-20T04:08:34+00:00",
                "reboot_reason": "Initiated by user - Reboot issued via NETCONF",
                "vdevice-host-name": "vm1",
                "vdevice-name": "172.16.255.11",
            },
            {
                "reboot_date_time-date": 164265171400,
                "lastupdated": 1642656685356,
                "vdevice-dataKey": "172.16.255.11-2022-01-20T04:08:34+00:00",
                "reboot_date_time": "2022-01-20T04:08:34+00:00",
                "reboot_reason": "Initiated by user - Reboot issued via NETCONF",
                "vdevice-host-name": "vm2",
                "vdevice-name": "172.16.255.12",
            },
        ]
        self.reboot_history_dataclass = [create_dataclass(Reboot, item) for item in self.reboot_history]
        self.device = [
            {  # vmanage
                "device-model": "vmanage",
                "deviceId": "1.1.1.1",
                "uuid": "aaaaaaaa-6169-445c-8e49-c0bdaaaaaaa",
                "cpuLoad": 4.71,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.1",
                "board-serial": "11122233",
                "personality": "vmanage",
                "memUsage": 57.0,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm200",
                "cpuState": "normal",
                "chassis-number": "aaaaaaaa-6aa9-445c-8e49-c0aaaaaaaaa9",
            }
        ]
        self.device_dataclass = create_dataclass(Device, self.device[0])
        self.wan_interfaces = [
            {
                'color': 'default',
                'vdevice-name': '1.1.1.1',
                'admin-state': 'up',
                'interface': 'eth1',
                'private-port': 12345,
                'vdevice-host-name': 'vm1',
                'public-ip': '1.1.1.1',
                'operation-state': 'up',
                'public-port': 12345,
                'private-ip': '1.1.1.1',
            },
            {
                'color': 'default',
                'vdevice-name': '1.1.1.1',
                'admin-state': 'up',
                'interface': 'eth1',
                'private-port': 11111,
                'vdevice-host-name': 'vm2',
                'public-ip': '1.1.1.1',
                'operation-state': 'up',
                'public-port': 11111,
                'private-ip': '1.1.1.1',
            },
        ]
        self.wan_interfaces_dataclass = [create_dataclass(WanInterface, item) for item in self.wan_interfaces]
        self.bfd_session = [
            {
                'src-ip': '1.1.1.1',
                'dst-ip': '1.1.1.2',
                'color': '3g',
                'system-ip': '1.1.1.1',
                'site-id': 1,
                'local-color': '3g',
                'state': 'up',
            },
            {
                'src-ip': '1.1.1.1',
                'dst-ip': '1.1.1.3',
                'color': '3g',
                'system-ip': '1.1.1.1',
                'site-id': 1,
                'local-color': '3g',
                'state': 'up',
            },
        ]
        self.bfd_session_dataclass = [create_dataclass(BfdSessionData, item) for item in self.bfd_session]
        self.bfd_session_down = [
            {
                'src-ip': '1.1.1.1',
                'dst-ip': '1.1.1.2',
                'color': '3g',
                'system-ip': '1.1.1.1',
                'site-id': 1,
                'local-color': '3g',
                'state': 'down',
            }
        ]
        self.device_unreachable = [
            {  # vmanage
                "device-model": "vmanage",
                "deviceId": "1.1.1.1",
                "uuid": "zzzzzzzz-6169-445c-8e49-c0bdaaaaaaa",
                "cpuLoad": 4.71,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.1",
                "board-serial": "11122233",
                "personality": "vmanage",
                "memUsage": 57.0,
                "reachability": "unreachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm200",
                "cpuState": "normal",
                "chassis-number": "aaaaaaaa-6aa9-445c-8e49-c0aaaaaaaaa9",
            }
        ]

    @patch('vmngclient.session.vManageSession')
    def test_get_device_crash_info(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.crash_info
        # Act
        answer = DeviceStateAPI(mock_session).get_device_crash_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.crash_info)

    @patch('vmngclient.session.vManageSession')
    def test_get_device_crash_info_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_device_crash_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.vManageSession')
    def test_get_device_control_connections_info(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.connections_info
        # Act
        answer = DeviceStateAPI(mock_session).get_device_control_connections_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.connections_info_dataclass)

    @patch('vmngclient.session.vManageSession')
    def test_get_device_control_connections_info_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_device_control_connections_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.vManageSession')
    def test_get_device_orchestrator_connections_info(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.connections_info
        # Act
        answer = DeviceStateAPI(mock_session).get_device_orchestrator_connections_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.connections_info_dataclass)

    @patch('vmngclient.session.vManageSession')
    def test_get_device_orchestrator_connections_info_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_device_orchestrator_connections_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.vManageSession')
    def test_get_device_reboot_history(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.reboot_history
        # Act
        answer = DeviceStateAPI(mock_session).get_device_reboot_history(device_id="1.1.1.11")
        # Assert
        self.assertEqual(answer, self.reboot_history_dataclass)

    @patch('vmngclient.session.vManageSession')
    def test_get_device_reboot_history_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_device_reboot_history(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.vManageSession')
    def test_get_system_status(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.device
        # Act
        answer = DeviceStateAPI(mock_session).get_system_status(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.device_dataclass)

    @patch('vmngclient.session.vManageSession')
    def test_get_system_status_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []

        # Act
        def answer():
            return DeviceStateAPI(mock_session).get_system_status(device_id="1.1.1.1")

        # Assert
        self.assertRaises(AssertionError, answer)

    @patch('vmngclient.session.vManageSession')
    def test_get_device_wan_interfaces(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.wan_interfaces
        # Act
        answer = DeviceStateAPI(mock_session).get_device_wan_interfaces(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.wan_interfaces_dataclass)

    @patch('vmngclient.session.vManageSession')
    def test_get_device_wan_interfaces_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_device_wan_interfaces(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    def test_get_colors(self):
        pass  # TODO fix method before test

    def test_enable_data_stream(self):
        pass  # TODO fix method before test

    @patch('vmngclient.session.vManageSession')
    def test_get_bfd_sessions(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.bfd_session
        # Act
        answer = DeviceStateAPI(mock_session).get_bfd_sessions(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.bfd_session_dataclass)

    @patch('vmngclient.session.vManageSession')
    def test_get_bfd_sessions_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_bfd_sessions(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    @patch('vmngclient.session.vManageSession')
    def test_wait_for_bfd_session_up(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.bfd_session
        # Act
        answer = DeviceStateAPI(mock_session).wait_for_bfd_session_up(system_ip="1.1.1.1")
        # Assert
        self.assertIsNone(answer)

    @patch('vmngclient.session.vManageSession')
    def test_wait_for_bfd_session_up_timeout(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.bfd_session_down

        # Act
        def answer():
            return DeviceStateAPI(mock_session).wait_for_bfd_session_up(
                system_ip="1.1.1.1", sleep_seconds=1, timeout_seconds=1
            )

        # Assert
        self.assertRaises(RetryError, answer)

    @patch('vmngclient.session.vManageSession')
    def test_wait_for_device_state(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.device
        # Act
        answer = DeviceStateAPI(mock_session).wait_for_device_state(device_id="1.1.1.1")
        # Assert
        self.assertTrue(answer)

    @patch('vmngclient.session.vManageSession')
    def test_wait_for_device_state_unreachable(self, mock_session):
        pass  # TODO fix method before test



import logging
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Set

from tenacity import retry, retry_if_result, stop_after_attempt, wait_fixed  # type: ignore

from vmngclient.dataclasses import AlarmData
from vmngclient.session import vManageSession
from vmngclient.utils.creation_tools import create_dataclass, flatten_dict

logger = logging.getLogger(__name__)


class AlarmLevel(Enum):
    minor = "Minor"
    major = "Major"
    critical = "Critical"


class Viewed(Enum):
    yes = "true"
    no = "false"


class AlarmsAPI:

    URL = '/dataservice/alarms'

    def __init__(self, session: vManageSession):
        self.session = session

    def get_alarms(
        self, hours: Optional[int] = None, level: Optional[str] = None, viewed: Optional[bool] = None
    ) -> List[AlarmData]:
        query: Dict[str, Any] = {"query": {"condition": "AND", "rules": []}}

        if hours:
            query["query"]["rules"].append(
                {"value": [str(hours)], "field": "entry_time", "type": "date", "operator": "last_n_hours"}
            )

        if level:
            query["query"]["rules"].append({"value": [level], "field": "severity", "type": "string", "operator": "in"})

        if viewed is not None:
            if viewed:
                value = Viewed.yes.value
            else:
                value = Viewed.no.value

            query["query"]["rules"].append(
                {"value": [value], "field": "acknowledged", "type": "bool", "operator": "equal"}
            )

        alarms = self.session.post(url=AlarmsAPI.URL, json=query).json()["data"]

        logger.info("Current alarms collected successfully.")

        return [create_dataclass(AlarmData, flatten_dict(alarm)) for alarm in alarms]

    def get_critical_alarms(self, hours: Optional[int] = None, viewed: Optional[bool] = None) -> List[AlarmData]:
        return self.get_alarms(hours, AlarmLevel.critical.value, viewed)

    def get_major_alarms(self, hours: Optional[int] = None, viewed: Optional[bool] = None) -> List[AlarmData]:
        return self.get_alarms(hours, AlarmLevel.major.value, viewed)

    def get_minor_alarms(self, hours: Optional[int] = None, viewed: Optional[bool] = None) -> List[AlarmData]:
        return self.get_alarms(hours, AlarmLevel.minor.value, viewed)

    def get_not_viewed_alarms(self, hours: Optional[int] = None) -> List[AlarmData]:
        return self.get_alarms(hours=hours, viewed=False)

    def mark_all_as_viewed(self) -> bool:
        """Marks all alarms as viewed.

        Returns:
          True if all alarms are viewed
        """

        self.session.post(f'{AlarmsAPI.URL}/markallasviewed')
        logger.info("Alarms mark as viewed.")

        return not self.get_not_viewed_alarms()

    def check_alarms(
        self, expected: List[Dict[str, str]], timeout_seconds: int = 240, sleep_seconds: int = 5
    ) -> Dict[str, set]:
        """Checks if alarms have occurred.

        For a specified period of time, the method queries the API about unread alarms and
        checks if they occur in the expected ones. If all alarms occur before the time expires,
        it will not poll any further.

        Args:
          expected(list): The list of expected alarms.
          timeout_seconds (int): Failure timeout.
          sleep_seconds (int): Sleep time.

        Returns:
          The dictionary with alarms that occurred (key 'found') and did not (key 'no-found')
        """

        alarms_expected = {create_dataclass(AlarmData, expected_alarm) for expected_alarm in expected}

        verification = AlarmVerification(logger, self.get_not_viewed_alarms)
        verification.verify(alarms_expected, timeout_seconds, sleep_seconds)

        logger.info(f"found alarms: {verification.found}")
        logger.info(f"not-found alarms: {verification.not_found}")

        return {'found': verification.found, 'not-found': verification.not_found}


class AlarmVerification:
    """The Alarms verification class

    Attributes:
        alarms_getter (Callable): The function that returns a set of current alarms.
        logger (Logger): Logger.
    """

    def __init__(self, logger: logging.Logger, alarms_getter: Callable[[], List[AlarmData]]) -> None:
        self.alarms_getter = alarms_getter
        self.logger = logger

    def __check(self, expected: AlarmData) -> bool:
        """The checking if the expected alarm is included in the alarm set.

        Args:
          expected(AlarmData): The expected alarm.

        Returns:
          True if the requested alarm appears in the current alarm set.
        """
        checked = {expected.lowercase().issubset(actual.lowercase()) for actual in self.alarms_getter()}
        return any(checked)

    def verify(self, expected: Set[AlarmData], timeout_seconds: int, sleep_seconds: int):
        """The verifying if the expected alarms is included in the actual alarms set.

        Args:
          expected(Set[AlarmData]): The set expected alarms.
          timeout_seconds (int): Failure timeout.
          sleep_seconds (int): Sleep time.

        """
        self.found: Set[AlarmData] = set()
        self.not_found: Set[AlarmData] = set()

        def _log_exception(retry_state):
            self.logger.error(f"Cannot found alarms in {timeout_seconds}.")
            self.logger.error(f"Original exception: {retry_state.outcome.exception()}.")

        def check(founds):
            return expected != founds

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check),
            retry_error_callback=_log_exception,
        )
        def wait_for_alarms():
            self.logger.info(f"waiting for alarms: {expected ^ self.found}")
            for expected_alarm in expected ^ self.found:
                if self.__check(expected_alarm):
                    self.found.add(expected_alarm)
            self.not_found = expected ^ self.found
            return self.found

        wait_for_alarms()



from typing import Optional
from urllib.parse import urljoin

import requests
from requests import PreparedRequest
from requests.auth import AuthBase
from requests.cookies import RequestsCookieJar


class InvalidCredentialsError(Exception):
    """Exception raised for invalid credentials.

    Attributes:
        username (str): vManage username.
        password (str): vManage password.
        message (str): precise error explanation.
    """

    def __init__(
        self, username: str, password: str, message: str = "Username and/or password is incorrect. Please try again!"
    ):
        self.username = username
        self.password = password
        self.message = message

    def __str__(self):
        return (
            f"Trying to access vManage with the following credentials: {self.username}/{self.password}. {self.message}"
        )


class vManageAuth(AuthBase):
    """Attaches vManage Authentication to the given Requests object.

    vManage REST API access control is based on sessions.
    The call method do whatever is required to make the vManage authentication work.
    The following are typical steps for a user to consume the API:
    1. Log in with a user name and password to establish a session.
    2. Get a cross-site request forgery prevention token, which is required for most POST operations.

    Attributes:
        base_url (str): url (with port if applicable) f.e. https://1.1.1.1:1111
        username (str): vManage username
        password (str): vManage user's password
        verify (bool): Controls whether we verify the server's TLS certificate.
                Defaults to True.
        expiration_time (int): Expiration token time in seconds.
                Defaults to None (unlimited).
        token (str): Access token

    """

    def __init__(self, base_url: str, username: str, password: str, verify: bool = False):
        self.base_url = base_url
        self.username = username
        self.password = password
        self.verify = verify  # TODO Handle `True` parameter
        self.expiration_time: Optional[int] = None  # Unlimited
        self.set_cookie: Optional[RequestsCookieJar] = None
        self.token: str = ""

    def get_cookie(self) -> RequestsCookieJar:
        """Check whether a user is successfully authenticated.

        If a user is successfully authenticated, the response body is empty and a valid session cookie is set.
        The response has entry in headers named `set-cookie` equal to JESSIONID={session hash}.
        If a user is un-authenticated, the response body contains a html login page with tag in it.

        Raises:
            InvalidCredentialsError: _description_

        Returns:
            RequestsCookieJar: _description_
        """
        security_payload = {
            'j_username': self.username,
            'j_password': self.password,
        }
        full_url = urljoin(self.base_url, "/j_security_check")
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        response = requests.post(
            url=full_url,
            data=security_payload,
            verify=self.verify,
            headers=headers,
        )
        if response.text != "":
            raise InvalidCredentialsError(self.username, self.password)
        return response.cookies

    def fetch_token(self, cookies: RequestsCookieJar) -> str:
        """If it is required, fetches vManage REST API token.

        The XSRF token is in the response body.
        XSRF token along with the JESSIONID cookie is used for ongoing API requests.

        Args:
            cookies (RequestsCookieJar): The JESSIONID={session hash} cookie is required to authenticate.

        Returns:
            str: Valid token.
        """
        full_url = urljoin(self.base_url, "/dataservice/client/token")
        headers = {'Content-Type': 'application/json'}
        response = requests.get(
            url=full_url,
            cookies=cookies,
            verify=self.verify,
            headers=headers,
        )
        return response.text

    def __call__(self, prepared_request: PreparedRequest) -> PreparedRequest:
        if self.expiration_time is None:
            if self.token == "":
                self.set_cookie = self.get_cookie()
                self.token = self.fetch_token(self.set_cookie)

        prepared_request.prepare_cookies(self.set_cookie)
        prepared_request.headers.update({"x-xsrf-token": self.token})
        return prepared_request



import datetime as dt
from typing import List, Optional

from attr import define, field  # type: ignore

from vmngclient.utils.creation_tools import FIELD_NAME, asdict, convert_attributes
from vmngclient.utils.device_model import DeviceModel
from vmngclient.utils.personality import Personality
from vmngclient.utils.reachability import Reachability


class DataclassBase:
    def __str__(self):
        return (
            f"{self.__class__.__name__}(\n"
            + "\n".join(f"    {attribute[0]}: {attribute[1]}," for attribute in asdict(self).items())
            + "\n)"
        )


@define(frozen=True, field_transformer=convert_attributes)
class AdminTech(DataclassBase):
    creation_time: dt.datetime = field(metadata={FIELD_NAME: "creationTime"})
    size: int
    filename: str = field(metadata={FIELD_NAME: "fileName"})
    state: str
    tac_state: Optional[str]
    device_ip: str = field(metadata={FIELD_NAME: "deviceIP"})
    system_ip: str = field(metadata={FIELD_NAME: "local-system-ip"})
    token_id: str = field(metadata={FIELD_NAME: "requestTokenId"})


@define(frozen=True, field_transformer=convert_attributes)
class DeviceAdminTech(DataclassBase):
    filename: str = field(metadata={FIELD_NAME: "fileName"})
    creation_time: dt.datetime = field(metadata={FIELD_NAME: "creationTime"})
    size: int
    state: str
    token_id: Optional[str] = field(default=None, metadata={FIELD_NAME: "requestTokenId"})


@define(frozen=True, field_transformer=convert_attributes)
class AlarmData(DataclassBase):
    component: Optional[str] = field(default=None)
    severity: Optional[str] = field(default=None)
    active: Optional[bool] = field(default=None)
    name: Optional[str] = field(default=None, metadata={FIELD_NAME: "type"})
    system_ip: Optional[str] = field(default=None, metadata={FIELD_NAME: "system-ip"})
    hostname: Optional[str] = field(default=None, metadata={FIELD_NAME: "host-name"})
    site_id: Optional[str] = field(default=None, metadata={FIELD_NAME: "site-id"})
    new_state: Optional[str] = field(default=None, metadata={FIELD_NAME: "new-state"})
    interface_name: Optional[str] = field(default=None, metadata={FIELD_NAME: "if-name"})
    vpn_id: Optional[str] = field(default=None, metadata={FIELD_NAME: "vpn-id"})

    def issubset(self, other: 'AlarmData') -> bool:
        field_keys = {field_key for field_key in self.__annotations__ if getattr(self, field_key)}
        return all(getattr(self, field_key) == getattr(other, field_key) for field_key in field_keys)

    def lowercase(self) -> 'AlarmData':
        data = dict()
        for field_key in self.__annotations__:
            attr = getattr(self, field_key)
            try:
                data[field_key] = attr.lower()
            except AttributeError:
                data[field_key] = attr

        return AlarmData(**data)


@define
class Device(DataclassBase):
    uuid: str
    status: str
    personality: Personality = field(converter=Personality)
    id: str = field(metadata={FIELD_NAME: "deviceId"})
    hostname: str = field(metadata={FIELD_NAME: "host-name"})
    reachability: Reachability = field(converter=Reachability)
    local_system_ip: str = field(metadata={FIELD_NAME: "local-system-ip"})
    memUsage: Optional[float] = field(default=None)
    mem_state: Optional[str] = field(default=None, metadata={FIELD_NAME: "memState"})
    cpu_state: Optional[str] = field(default=None, metadata={FIELD_NAME: "cpuState"})
    cpu_load: Optional[float] = field(default=None, metadata={FIELD_NAME: "cpuLoad"})
    state_description: Optional[str] = field(default=None)
    connected_vManages: List[str] = field(factory=list, metadata={FIELD_NAME: "connectedVManages"})
    model: Optional[str] = field(default=None, metadata={FIELD_NAME: "device-model"})
    board_serial: Optional[str] = field(default=None, metadata={FIELD_NAME: 'board-serial'})
    vedgeCertificateState: Optional[str] = field(default=None, metadata={FIELD_NAME: 'vedgeCertificateState'})  # TODO
    chasis_number: Optional[str] = field(default=None, metadata={FIELD_NAME: 'chasisNumber'})

    @property
    def is_reachable(self) -> bool:
        return self.reachability is Reachability.REACHABLE


@define(field_transformer=convert_attributes)
class Reboot(DataclassBase):
    reason: str = field(metadata={FIELD_NAME: "reboot_reason"})
    dateTime: dt.datetime = field(metadata={FIELD_NAME: "reboot_date_time"})
    vdeviceName: str = field(metadata={FIELD_NAME: "vdevice-name"})
    rebootDateTimeDate: dt.datetime = field(metadata={FIELD_NAME: "reboot_date_time-date"})
    vdeviceDataKey: str = field(metadata={FIELD_NAME: "vdevice-dataKey"})
    lastUpdated: dt.datetime = field(metadata={FIELD_NAME: "lastupdated"})
    vdeviceHostName: int = field(metadata={FIELD_NAME: "vdevice-host-name"})


@define
class WanInterface(DataclassBase):
    color: str
    vDeviceIp: str = field(metadata={FIELD_NAME: 'vdevice-name'})
    vDeviceName: str = field(metadata={FIELD_NAME: 'vdevice-host-name'})
    adminState: str = field(metadata={FIELD_NAME: 'admin-state'})
    interfaceName: str = field(metadata={FIELD_NAME: 'interface'})
    privateIp: str = field(metadata={FIELD_NAME: 'private-ip'})
    publicIp: str = field(metadata={FIELD_NAME: 'public-ip'})
    privatePort: int = field(metadata={FIELD_NAME: 'private-port'})
    publicPort: int = field(metadata={FIELD_NAME: 'public-port'})
    operationalState: str = field(metadata={FIELD_NAME: 'operation-state'})


@define
class Connection(DataclassBase):
    state: str
    peerType: str = field(metadata={FIELD_NAME: 'peer-type'})
    systemIp: str = field(metadata={FIELD_NAME: 'system-ip'})


@define
class BfdSessionData(DataclassBase):
    state: str
    siteId: str = field(metadata={FIELD_NAME: "site-id"})
    sourceTlocColor: str = field(metadata={FIELD_NAME: "local-color"})
    remoteTlocColor: str = field(metadata={FIELD_NAME: "color"})
    deviceIp: str = field(metadata={FIELD_NAME: "system-ip"})
    sourceIp: str = field(metadata={FIELD_NAME: "src-ip"})
    destinationPublicIp: str = field(metadata={FIELD_NAME: "dst-ip"})


@define
class OmpPeerData(DataclassBase):
    type: str
    state: str
    peerIp: str = field(metadata={FIELD_NAME: "peer"})
    siteId: str = field(metadata={FIELD_NAME: "site-id"})
    domainId: str = field(metadata={FIELD_NAME: "domain-id"})
    deviceIp: str = field(metadata={FIELD_NAME: "vdevice-name"})


@define
class OmpReceivedRouteData(DataclassBase):
    protocol: str
    peerIp: str = field(metadata={FIELD_NAME: "from-peer"})


@define
class OmpAdvertisedRouteData(DataclassBase):
    protocol: str
    peerIp: str = field(metadata={FIELD_NAME: "to-peer"})


@define
class OmpReceivedTlocData(DataclassBase):
    peerIp: str = field(metadata={FIELD_NAME: "from-peer"})


@define
class OmpAdvertisedTlocData(DataclassBase):
    peerIp: str = field(metadata={FIELD_NAME: "to-peer"})


@define
class OmpServiceData(DataclassBase):
    name: str = field(metadata={FIELD_NAME: "service"})
    status: Optional[str] = field(default=None)


@define
class OmpSummaryData(DataclassBase):
    oper_state: str = field(metadata={FIELD_NAME: "operstate"})
    admin_state: str = field(metadata={FIELD_NAME: "adminstate"})
    routes_received: str = field(metadata={FIELD_NAME: "routes-received"})
    tlocs_received: str = field(metadata={FIELD_NAME: "tlocs-received"})
    routes_sent: str = field(metadata={FIELD_NAME: "routes-sent"})
    tlocs_installed: str = field(metadata={FIELD_NAME: "tlocs-installed"})
    tlocs_sent: str = field(metadata={FIELD_NAME: "tlocs-sent"})
    vsmart_peers: str = field(metadata={FIELD_NAME: "vsmart-peers"})


@define
class EventData(DataclassBase):
    system_ip: str
    vmanage_system_ip: str
    tenant: str
    device_type: str
    component: str
    severity_level: str
    host_name: str
    event: str
    details: str
    event_name: str = field(metadata={FIELD_NAME: "eventname"})


@define(frozen=True)
class User(DataclassBase):
    group: List[str]
    username: str = field(metadata={FIELD_NAME: "userName"})
    locale: Optional[str] = field(default=None)
    password: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)
    resource_group: Optional[str] = field(default=None, metadata={FIELD_NAME: "resGroupName"})


@define
class Template(DataclassBase):
    device_type_str: str = field(metadata={FIELD_NAME: "deviceType"})
    device_type: DeviceModel = field(init=False)
    last_updated_by: str = field(metadata={FIELD_NAME: "lastUpdatedBy"})
    resource_group: str = field(metadata={FIELD_NAME: "resourceGroup"})
    template_class: str = field(metadata={FIELD_NAME: "templateClass"})
    config_type: str = field(metadata={FIELD_NAME: "configType"})
    id: str = field(metadata={FIELD_NAME: "templateId"})
    factory_default: bool = field(metadata={FIELD_NAME: "factoryDefault"})
    name: str = field(metadata={FIELD_NAME: "templateName"})
    devices_attached: int = field(metadata={FIELD_NAME: "devicesAttached"})
    description: str = field(metadata={FIELD_NAME: "templateDescription"})
    draft_mode: str = field(metadata={FIELD_NAME: "draftMode"})
    last_updated_on: dt.datetime = field(metadata={FIELD_NAME: "lastUpdatedOn"})
    template_attached: int = field(metadata={FIELD_NAME: "templateAttached"})

    def __attrs_post_init__(self):
        self.device_type = DeviceModel(self.device_type_str)


@define
class Speedtest(DataclassBase):
    device_ip: str
    device_name: str
    destination_ip: str
    destination_name: str
    status: str
    up_speed: float
    down_speed: float


@define(frozen=True)
class PacketSetup(DataclassBase):
    session_id: str = field(metadata={FIELD_NAME: "sessionId"})
    is_new_session: bool = field(metadata={FIELD_NAME: "isNewSession"})


@define(frozen=True)
class Status(DataclassBase):
    file_download_status: Optional[str] = field(default=None, metadata={FIELD_NAME: "fileDownloadStatus"})
    file_size: Optional[int] = field(default=None, metadata={FIELD_NAME: "fileSize"})


@define(frozen=True)
class ServiceConfigurationData(DataclassBase):
    """Administration -> Service Configuration"""

    vmanage_id: str = field(metadata={FIELD_NAME: "vmanageID"})
    device_ip: str = field(metadata={FIELD_NAME: "deviceIP"})  # consider using ip4 module to verify
    services: dict = field(metadata={FIELD_NAME: "services"})  # consider using nested dataclasses
    persona: str = field(default="COMPUTE_AND_DATA")  # TODO Enum
    username: Optional[str] = field(default=None)
    password: Optional[str] = field(default=None)


@define(frozen=True)
class CloudConnectorData(DataclassBase):
    """Administration -> Settings -> SD-AVC Cloud Connector"""

    client_id: str = field(metadata={FIELD_NAME: "clientId"})
    client_secret: str = field(metadata={FIELD_NAME: "clientSecret"})
    org_name: str = field(metadata={FIELD_NAME: "orgName"})
    telemetry_enabled: bool = field(metadata={FIELD_NAME: "telemetryEnabled"})
    affinity: Optional[str] = field(default=None)
    cloud_enabled: bool = field(default=True, metadata={FIELD_NAME: "cloudEnabled"})


@define(frozen=True)
class CloudServicesSettings(DataclassBase):
    """Administration -> Settings -> Cloud Services"""

    enabled: bool = field(metadata={FIELD_NAME: "enabled"})
    otp: Optional[str] = field(default=None, metadata={FIELD_NAME: "otp"})
    cloud_gateway_url: Optional[str] = field(default=None, metadata={FIELD_NAME: "cloudGatewayUrl"})


@define(frozen=True)
class CloudOnRampForSaasMode(DataclassBase):
    """
    Administration -> Settings -> Cloud on Ramp for Saas
    """

    mode: str = field(metadata={FIELD_NAME: "mode"})



import logging
import logging.config
from pathlib import Path
from typing import Final

LOGGING_CONF_DIR: Final[str] = str(Path(__file__).parents[0] / 'logging.conf')

vmngclient_logger = logging.getLogger(__name__)

if not vmngclient_logger.handlers:
    logging.config.fileConfig(LOGGING_CONF_DIR, disable_existing_loggers=False)



import logging
from typing import List, Union, cast

from vmngclient.dataclasses import CloudConnectorData, CloudServicesSettings, ServiceConfigurationData, User
from vmngclient.session import vManageSession
from vmngclient.utils.creation_tools import asdict, create_dataclass

logger = logging.getLogger(__name__)


class UserAlreadyExistsError(Exception):
    pass


class UserDoesNotExists(Exception):
    pass


class UsersAPI:
    def __init__(self, session: vManageSession) -> None:
        self.session = session

    def get_all_users(self) -> List[User]:
        url_path = "/dataservice/admin/user"
        users = self.session.get_data(url_path)
        logger.debug(f"List of users: {users}")
        return [create_dataclass(User, user) for user in users]

    def exists(self, username: str) -> bool:
        return username in [user.username for user in self.get_all_users()]

    def create_user(self, user: User) -> None:
        if self.exists(user.username):
            raise UserAlreadyExistsError(f"{user.username} already exists.")
        url_path = "/dataservice/admin/user"
        data = asdict(user)  # type: ignore

        response = self.session.post(url=url_path, json=data)
        logger.info(response)

    def delete_user(self, username: str) -> bool:
        if not self.exists(username):
            raise UserDoesNotExists(f"{username} does not exists.")
        url_path = f"/dataservice/admin/user/{username}"
        logger.debug(f"Deleting user {username}.")
        response = self.session.delete(url_path)
        return True if response.status_code == 200 else False


class ClusterManagementAPI:
    """Covers clusterManagement API calls.

    Attributes:
        session: logged in API admin session
    """

    def __init__(self, session: vManageSession) -> None:
        self.session = session

    def modify_cluster_setup(self, service_configuration: ServiceConfigurationData) -> bool:
        """Updates vManage cluster configuration.

        Args:
            service_configuration: vManage cluster config
        """
        url_path = "/dataservice/clusterManagement/setup"
        data = asdict(service_configuration)  # type: ignore
        response = self.session.put(url_path, data)
        return True if response.status_code == 200 else False

    def get_cluster_management_health_status(self) -> Union[dict, list]:
        """Gets Cluster Management Service Reachability health status.

        Returns:
            TODO return health status dataclass
        """
        url_path = "/dataservice/clusterManagement/health/status"
        return self.session.get_data(url_path)


class AdministrationSettingsAPI:
    def __init__(self, session: vManageSession) -> None:
        """Covers Administration Settings API calls.

        Args:
            session: logged in API admin session
        """
        self.session = session

    def get_sdavc_cloud_connector_config(self) -> CloudConnectorData:
        """Gets SD-AVC Cloud Connector Config.

        Returns:
            CloudConnector dataclass config
        """
        url_path = "/dataservice/sdavc/cloudconnector"
        response = cast(dict, self.session.get_json(url_path))
        return create_dataclass(CloudConnectorData, response)

    def enable_sdavc_cloud_connector(self, cloud_connector: CloudConnectorData) -> bool:
        """Enables SD-AVC Cloud Connector on vManage."""
        url_path = "/dataservice/sdavc/cloudconnector"
        data = asdict(cloud_connector)  # type: ignore
        response = self.session.post(url_path, json=data)
        return True if response.status_code == 200 else False

    def disable_sdavc_cloud_connector(self) -> bool:
        """Disables SD-AVC Cloud Connector on vManage."""
        url_path = "/dataservice/sdavc/cloudconnector"
        data = {"cloudEnabled": False}
        response = self.session.put(url_path, data)
        return True if response.status_code == 200 else False

    def get_cloud_services(self) -> CloudServicesSettings:
        url_path = "/dataservice/settings/configuration/cloudservices"
        data = self.session.get_data(url_path)[0]
        return create_dataclass(CloudServicesSettings, data)

    def set_cloud_services(self, config: CloudServicesSettings) -> bool:
        url_path = "/dataservice/settings/configuration/cloudservices"
        response = self.session.post(url_path, asdict(config))  # type: ignore
        return True if response.status_code == 200 else False

    def get_cloud_on_ramp_for_saas_mode(self):
        """Get information about Cloud on Ramp for Saas mode"""
        url_path = "/dataservice/settings/configuration/cloudx"
        return self.session.get_data(url_path)

    def enable_cloud_on_ramp_for_saas_mode(self, disable=False):
        """Enable or disable COR for SaaS"""
        url_path = "/dataservice/settings/configuration/cloudx"
        data = {"mode": "on"} if not disable else {"mode": "off"}
        response = self.session.put(url_path, data)
        return True if response.status == 200 else False



import unittest
from unittest import TestCase, mock

from vmngclient.vmanage_auth import InvalidCredentialsError, vManageAuth


class MockResponse:
    def __init__(self, status_code: int, text: str):
        self._status_code = status_code
        self._text = text

    def cookies(self) -> str:  # TODO
        return "JSESSIONID=xyz"

    @property
    def status_code(self) -> int:
        return self._status_code

    @property
    def text(self) -> str:  # TODO
        return self._text


def mocked_requests_method(*args, **kwargs):
    url_response = {
        "https://1.1.1.1:1111/j_security_check": {
            "admin": MockResponse(200, ""),
            "invalid_username": MockResponse(200, "<html>error</html>"),
        }
    }

    full_url = kwargs.get('url', "")
    data = kwargs.get('data', {})
    if full_url in url_response:
        return url_response[full_url][data['j_username']]

    return MockResponse(404, "error")


class TestvManageAuth(TestCase):
    def setUp(self):
        self.base_url = "https://1.1.1.1:1111"
        self.password = "admin"

    @mock.patch('requests.post', side_effect=mocked_requests_method)
    def test_get_cookie(self, mock_post):
        # Arrange
        username = "admin"
        security_payload = {
            'j_username': username,
            'j_password': "admin",
        }
        auth = vManageAuth(self.base_url, username, self.password)
        # Act
        auth.get_cookie()

        # Assert
        mock_post.assert_called_with(
            url="https://1.1.1.1:1111/j_security_check",
            data=security_payload,
            verify=False,
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
        )

    @mock.patch('requests.post', side_effect=mocked_requests_method)
    def test_get_cookie_invalid_username(self, mock_post):
        # Arrange
        username = "invalid_username"
        security_payload = {
            'j_username': username,
            'j_password': "admin",
        }
        auth = vManageAuth(self.base_url, username, self.password)
        # Act
        with self.assertRaises(InvalidCredentialsError):
            auth.get_cookie()

        # Assert
        mock_post.assert_called_with(
            url="https://1.1.1.1:1111/j_security_check",
            data=security_payload,
            verify=False,
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
        )

    @mock.patch('requests.cookies.RequestsCookieJar')
    @mock.patch('requests.get', side_effect=mocked_requests_method)
    def test_fetch_token(self, mock_get, cookies):
        # Arrange
        valid_url = "https://1.1.1.1:1111/dataservice/client/token"
        auth = vManageAuth(self.base_url, "admin", self.password)

        # Act
        auth.fetch_token(cookies)

        # Assert
        mock_get.assert_called_with(
            url=valid_url, verify=False, headers={'Content-Type': 'application/json'}, cookies=cookies
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vmngclient', 'vmngclient.api', 'vmngclient.tests', 'vmngclient.utils']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'aiohttp>=3.8.1,<4.0.0',
 'attrs>=21.4.0,<22.0.0',
 'ciscoconfparse>=1.6.40,<2.0.0',
 'parameterized>=0.8.1,<0.9.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.27.1,<3.0.0',
 'tenacity>=8.1.0,<9.0.0']

setup_kwargs = {
    'name': 'vmngclient',
    'version': '0.2.0',
    'description': 'Universal vManage API',
    'long_description': '# vManage-client\n[![Python3.8](https://img.shields.io/static/v1?label=Python&logo=Python&color=3776AB&message=3.8)](https://www.python.org/)\n\nvManage client is a package for creating simple and parallel automatic requests via official vManageAPI. It is intended to serve as a multiple session handler (provider, provider as a tenant, tenant). The library is not dependent on environment which is being run, you just need a connection to any vManage.\n\n## Installation\n```console\npip install vmngclient\n```\n\n## Hello world example\n\n<details>\n    <summary>Python (click to expand)</summary>\n\n```Python\nfrom vmngclient.session import create_vManageSession\n\n\nbase_url = "sandbox-sdwan-2.cisco.com/"\nusername = "devnetuser"\npassword = "RG!_Yw919_83"\nsession = create_vManageSession(url=base_url, username=username, password=password)\n\n\n>>> "Logged as devnetuser. The session type is SessionType.TENANT"\n>>> {\'title\': \'Cisco vManage\', \'version\': \'20.4.2.1\', \'applicationVersion\': \'20.4R-vbamboo-16-Dec-2021 19:07:17 PST\', \'applicationServer\': \'vmanage\', \'copyright\': \'Copyright (c) 2022, Cisco. All rights reserved.\', \'time\': \'2022-12-01 13:45:44\', \'timeZone\': \'UTC\', \'logo\': \'/dataservice/client/logo.png\'}\n```\n</details>\n\n### Note:\nTo remove `InsecureRequestWarning`, you can include in your scripts:\n```Python\nimport urllib3\nurllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)\n```\n\n## User creation example\n\n<details>\n    <summary>Python (click to expand)</summary>\n\n```Python\nfrom vmngclient.api.administration import UserAlreadyExistsError, UserApi\nfrom vmngclient.dataclasses import User\nfrom vmngclient.session import create_vManageSession\n\nsession = create_vManageSession(url=..., username=..., password=...)\nuser_api = UserApi(session)\n\ntest_user = User(\n    group=["basic"],\n    description="Demo User",\n    username="demouser",\n    password="password",\n    locale="en_US",\n    resource_group="global"\n)\n\ntry:\n    user_api.create_user(test_user)\nexcept UserAlreadyExistsError as error:\n    print(f"User {username} already exists.")\n```\n</details>\n\n## API usage examples\n\n### AdminTechAPI\n\n<details>\n    <summary>Python (click to expand)</summary>\n\n```Python\nfrom vmngclient.session import create_vManageSession\nfrom vmngclient.api.admin_tech_api import AdminTechAPI\n\nsession = create_vManageSession(url=..., username=..., password=...)\nadmintech = AdminTechAPI(session)\nfilename = admintech.generate("172.16.255.11")\nadmintech.download(filename)\nadmintech.delete(filename)\n```\n\n</details>\n\n\n## Contributing, reporting issues, seeking support\nPlease contact authors direcly or via Issues Github page.\n\n## **Enviroment setup**\n1. Download Python3.8 or higher.\n2. Download repository\n    ```\n    git clone https://github.com/CiscoDevNet/vManage-client.git\n    ```\n3. Install poetry v1.1.13\n    ```\n    pip install poetry==1.1.13\n    ```\n4. Install dependecies \n    ```\n    poetry install\n    ```\n5. Activate `pre-commit`\n    ```\n    pre-commit install\n    ```\n\n## **Add new feature**\nTo add new feature create new branch and implement it. Before making a pull request make sure that `pre-commit` passes.\n- **Building package for tests**\\\n    To make a `.whl` file run\n    ```\n    poetry build\n    ```\n    Then in `/vManage-client/dist/` directory there is a `.whl` file named `vmngclient-<version>-py3-none-any.whl`, which can be installed by running\n    ```\n    pip install vmngclient-<version>-py3-none-any.whl\n    ```',
    'author': 'kagorski',
    'author_email': 'kagorski@cisco.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CiscoDevNet/vManage-client',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Methods covering essential API endpoints and related data classes."""
import json
from contextlib import contextmanager
from enum import Enum
from typing import Iterator, List, Union

from tenacity import retry, retry_if_result, stop_after_attempt, wait_fixed  # type: ignore

from vmngclient.dataclasses import BfdSessionData, Connection, Device, Reboot, WanInterface
from vmngclient.session import vManageSession
from vmngclient.utils.creation_tools import create_dataclass
from vmngclient.utils.operation_status import OperationStatus
from vmngclient.utils.personality import Personality
from vmngclient.utils.reachability import Reachability


# TODO link that with dataclass
class DeviceField(Enum):
    HOSTNAME = 'hostname'
    ID = 'id'


class DeviceNotFoundError(Exception):
    pass


class DevicesAPI:
    """API methods of vManage for getting devices and controllers.

    Attributes:
        session: logged in API client session
        devices: List with system status for all devices
    """

    def __init__(self, session: vManageSession) -> None:
        self.session = session

    def __str__(self) -> str:
        return str(self.session)

    @property
    def controllers(self) -> List[Device]:
        """List of controller devices only."""
        return [
            controller
            for controller in self.devices
            if controller.personality in [Personality.VMANAGE, Personality.VSMART]
        ]

    @property
    def orchestrators(self) -> List[Device]:
        """List of orchestrator devices only."""
        return [orchestrator for orchestrator in self.devices if Personality.VBOND is orchestrator.personality]

    @property
    def edges(self) -> List[Device]:
        """List of edge devices only."""
        return [edge for edge in self.devices if Personality.EDGE is edge.personality]

    @property
    def vsmarts(self) -> List[Device]:
        """List of vsmart devices only."""
        return [vsmart for vsmart in self.devices if Personality.VSMART is vsmart.personality]

    @property
    def system_ips(self) -> List[str]:
        """List of device system IP addresses."""
        return [device.local_system_ip for device in self.devices]

    @property
    def ips(self):
        """List of device IP addresses."""
        return [device.id for device in self.devices]

    def get_system_ip_based_on_local_system_ip(self, local_system_ip) -> str:
        for dev in self.devices:
            if local_system_ip == dev.local_system_ip:
                return dev.id
        return ''

    @property
    def devices(self) -> List[Device]:
        """List of all devices."""
        devices_basic_info = self.session.get_data('/dataservice/device')

        devices_ids = ""
        for device in devices_basic_info:
            devices_ids += f"&deviceId={device['deviceId']}"

        devices_full_info = self.session.get_data(f'/dataservice/device/system/info?{devices_ids}')

        return [create_dataclass(Device, device) for device in devices_full_info]

    def get_device_details(self, uuid: str) -> Device:
        """Gets system information for a device.

        Args:
            device_id: device ID (usually system-ip)

        Returns:
            Device object
        """
        devices = self.session.get_data(f'/dataservice/system/device/vedges?uuid={uuid}')

        assert len(devices) == 1, 'Expected system info response list to have one member'

        return create_dataclass(Device, devices[0])

    def count_devices(self, personality: Personality) -> int:
        """Gets number of devices of given personality.

        Args:
            personality: personality of the device

        Returns:
            count of devices
        """
        return sum([1 for device in self.devices if device.personality == personality.value])

    def get_tenants(self) -> Union[list, dict]:
        """Gets Tenants.

        Returns:
            Tenants
        """
        tenants = self.session.get_data('/dataservice/tenant')

        return tenants

    def get_reachable_devices(self, personality: Personality):
        """Get reachable devices by personality.

        Args:
            personality: personality of the device

        Returns:
            reachable devices
        """
        unsupported_personality = [Personality.VMANAGE]
        assert personality not in unsupported_personality, 'Unsupported personality for reachable endpoint'
        return self.session.get_data(f'/dataservice/device/reachable?personality={personality.value}')

    def send_certificate_state_to_controllers(
        self,
        sleep_seconds: int = 5,
        timeout_seconds: int = 600,
    ) -> bool:
        """Sending the current status of certificates to the controllers

        Returns:
            bool: True if all ok, False like something wrong
        """

        def _log_exception(retry_state):
            self.session.__get_logger(f"Orignial exception: {retry_state.outcome.exception()}.")

        def check_state(action_data):
            list_action = [action['status'] == OperationStatus.SUCCESS.value for action in action_data]
            return not all(list_action)

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_state),
            retry_error_callback=_log_exception,
        )
        def wait_for_state():
            status_api = f'/dataservice/device/action/status/{action_id}'
            return self.session.get_data(f'{status_api}')

        response = self.session.post('/dataservice/certificate/vedge/list?action=push').json()
        if response.get('id'):
            action_id = response['id']
        else:
            raise FailedSend('Failed to push edges list certificates')

        return True if wait_for_state() else False

    def get(self, field: DeviceField, value: str) -> Device:
        supported_fields = [
            DeviceField.HOSTNAME,
            DeviceField.ID,
        ]

        if field not in supported_fields:
            raise TypeError(f"{field} is not supported. Available fields: {supported_fields}")

        for device in self.devices:
            if getattr(device, field.value) == value:
                return device
        raise DeviceNotFoundError(f"Device with `{field.value}` equals to `{value}` does not exists.")


class DeviceStateAPI:
    """Basic API methods of vManage.

    Attributes:
        session: logged in API client session
    """

    def __init__(self, session: vManageSession) -> None:
        self.session = session

    def __str__(self) -> str:
        return str(self.session)

    def get_device_crash_info(self, device_id: str) -> Union[list, dict]:
        """Gets crash info for a device.

        Args:
             device_id: device ID (usually system-ip)

        Returns:
            Union[list, dict]: list of dicts (FIXME: add mapping to a NamedTuples)
        """
        return self.session.get_data(f'/dataservice/device/crashlog?deviceId={device_id}')

    def get_device_control_connections_info(self, device_id) -> List[Connection]:
        """Gets control connections for a device.

        Args:
            device_id: device ID (usually system-ip)

        Returns:
            list of Connection objects
        """
        items = self.session.get_data(f'/dataservice/device/control/connections?deviceId={device_id}')

        return [create_dataclass(Connection, item) for item in items]

    def get_device_orchestrator_connections_info(self, device_id) -> List[Connection]:
        """Gets orchestrator connections for a device

        Args:
            device_id: device ID (usually system-ip)

        Returns:
            list of Connection objects
        """
        items = self.session.get_data(f'/dataservice/device/orchestrator/connections?deviceId={device_id}')

        return [create_dataclass(Connection, item) for item in items]

    def get_device_reboot_history(self, device_id):
        """Gets device reboots list.

        Args:
            device_id: device ID (usually system-ip)

        Returns:
            list of Reboot objects
        """
        items = self.session.get_data(f'/dataservice/device/reboothistory?deviceId={device_id}')

        return [create_dataclass(Reboot, item) for item in items]

    def get_system_status(self, device_id: str) -> Device:
        """Get system information for a device.

        Args:
            device_id: device ID (usually system-ip)

        Returns:
           Device object
        """
        devices = self.session.get_data(f'/dataservice/device/system/info?deviceId={device_id}')

        assert len(devices) == 1, 'Expected system info response list to have one member'

        return create_dataclass(Device, devices[0])

    def get_device_wan_interfaces(self, device_id: str):
        wan_interfaces = self.session.get_data(f'/dataservice/device/control/waninterface?deviceId={device_id}')
        return [create_dataclass(WanInterface, wan_ifc) for wan_ifc in wan_interfaces]

    def get_colors(self, device_id: str) -> List[str]:
        url = '/dataservice/device/bfd/state/device/tlocInterfaceMap'
        colors_raw = DevicesAPI(self.session).session.get_data(url + f'?deviceId={device_id}')
        json_colors = json.loads(str(colors_raw.read(), 'utf-8'))
        colors = list(json_colors["intfList"].keys())

        return colors

    @contextmanager
    def enable_data_stream(self) -> Iterator:  # TODO check
        try:
            url_path = "/dataservice/settings/configuration/vmanagedatastream"
            data_stream_status = self.session.get_data(url_path)[0]
            query = {  # TODO Dict[str, obj]
                "enable": "True",
                "ipType": "systemIp",
                "serverHostName": "systemIp",
                "vpn": "0",
            }
            url_path = "/dataservice/settings/configuration/vmanagedatastream"
            self.session.post(url=url_path, json=query)
            yield None
        finally:
            url_path = "/dataservice/settings/configuration/vmanagedatastream"
            self.session.post(url=url_path, json=data_stream_status)

    def get_bfd_sessions(self, device_id: str) -> List[BfdSessionData]:
        items = self.session.get_data(f'/dataservice/device/bfd/sessions?deviceId={device_id}')

        return [create_dataclass(BfdSessionData, item) for item in items]

    def wait_for_bfd_session_up(
        self, system_ip: str, sleep_seconds: int = 5, timeout_seconds: int = 60, exp_state: str = 'up'
    ):
        def check_state(bfd_sessions):
            return not all([bfd_session.state == exp_state for bfd_session in bfd_sessions])

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_state),
        )
        def wait_for_bfd_session_come_up():
            return self.get_bfd_sessions(system_ip)
            # from my observation it is necessary to wait minimum 5 seconds for BFD's session gets up state

        wait_for_bfd_session_come_up()

    def wait_for_device_state(
        self,
        device_id: str,
        sleep_seconds: int = 5,
        timeout_seconds: int = 600,
        exp_state: Reachability = Reachability.REACHABLE,
    ):
        """
        Waiting for the state of the machine.

        Args:
          device_id(Str): Device ID (usually system-ip)
          timeout_seconds(int): Failure timeout.
          sleep_seconds(int): Sleep time.
          exp_state(Reachability): The expected state of the machine

        Returns:
          True if the expected state has been achieved

        """

        def _log_exception(retry_state):
            self.session.__get_logger(f"Orignial exception: {retry_state.outcome.exception()}.")

        def check_state(state):
            return state == exp_state.value

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_state),
            retry_error_callback=_log_exception,
        )
        def wait_for_state():
            return self.get_system_status(device_id).reachability

        return True if wait_for_state() else False


class FailedSend(Exception):
    """Used when a referenced item is not found"""

    pass


__all__ = ['Device']



import unittest
from typing import Any
from unittest.mock import patch

from parameterized import parameterized  # type: ignore

from vmngclient.utils.response import (
    VManageResponseException,
    get_json_data,
    get_json_data_as_dict,
    get_json_data_as_list,
)


class TestResponse(unittest.TestCase):
    @parameterized.expand(
        [
            (True, None),
            (True, "string"),
            (True, {}),
            (True, {"key1": "string", "key2": 12, "key3": {"nested": None}}),
            (True, 15),
            (False, {"data": None}),
            (False, {"data": {}}),
            (False, {"data": []}),
            (False, {"data": "something"}),
            (False, {"data": [1, 2, 3]}),
            (False, {"data": {"key1": "string", "key2": 12, "key3": None}}),
            (False, {"data": [{"key1": "string"}, {"key2": {}, "key3": None}, "data"]}),
        ]
    )
    @patch('requests.Response')
    def test_get_json_data(self, raises: bool, json: Any, mock_response):
        mock_response.json.return_value = json
        if not raises:
            self.assertEqual(get_json_data(mock_response), json["data"])
        else:
            with self.assertRaises(VManageResponseException):
                get_json_data(mock_response)

    @parameterized.expand(
        [
            (True, None),
            (True, "string"),
            (True, {}),
            (True, {"key1": "string", "key2": 12, "key3": {"nested": None}}),
            (True, 15),
            (True, {"data": None}),
            (True, {"data": {}}),
            (False, {"data": []}),
            (True, {"data": "something"}),
            (False, {"data": [1, 2, 3]}),
            (True, {"data": {"key1": "string", "key2": 12, "key3": None}}),
            (False, {"data": [{"key1": "string"}, {"key2": {}, "key3": None}, "data"]}),
        ]
    )
    @patch('requests.Response')
    def test_get_json_data_as_list(self, raises: bool, json: Any, mock_response):
        mock_response.json.return_value = json
        if not raises:
            self.assertEqual(get_json_data_as_list(mock_response), json["data"])
        else:
            with self.assertRaises(VManageResponseException):
                get_json_data_as_list(mock_response)

    @parameterized.expand(
        [
            (True, None),
            (True, "string"),
            (True, {}),
            (True, {"key1": "string", "key2": 12, "key3": {"nested": None}}),
            (True, 15),
            (True, {"data": None}),
            (False, {"data": {}}),
            (True, {"data": []}),
            (True, {"data": "something"}),
            (True, {"data": [1, 2, 3]}),
            (False, {"data": {"key1": "string", "key2": 12, "key3": None}}),
            (True, {"data": [{"key1": "string"}, {"key2": {}, "key3": None}, "data"]}),
        ]
    )
    @patch('requests.Response')
    def test_get_json_data_as_dict(self, raises: bool, json: Any, mock_response):
        mock_response.json.return_value = json
        if not raises:
            self.assertEqual(get_json_data_as_dict(mock_response), json["data"])
        else:
            with self.assertRaises(VManageResponseException):
                get_json_data_as_dict(mock_response)
